# -*- coding: utf-8 -*-
#
# File: adapters.py
#
# GNU General Public License (GPL)
#

from collective.behavior.talcondition.utils import _evaluateExpression
from ftw.labels.interfaces import ILabelJar
from ftw.labels.labeling import ANNOTATION_KEY as FTW_LABELS_ANNOTATION_KEY
from ftw.labels.labeling import Labeling
from imio.helpers.cache import get_current_user_id
from imio.helpers.cache import get_plone_groups_for_user
from imio.helpers.cache import invalidate_cachekey_volatile_for
from plone import api
from plone.app.querystring.queryparser import parseFormquery
from plone.memoize import ram
from Products.CMFPlone.utils import safe_unicode
from Products.PloneMeeting.config import ITEM_LABELS_ACCESS_CACHE_ATTR
from Products.PloneMeeting.config import PMMessageFactory as _
from Products.PloneMeeting.utils import _base_extra_expr_ctx
from zope.annotation import IAnnotations


def ftw_labels_jar_discovery(context):
    """Return the root where ftw.labels are defined, here the MeetingConfig."""
    tool = api.portal.get_tool('portal_plonemeeting')
    cfg = tool.getMeetingConfig(context)
    return ILabelJar(cfg)


class PMLabeling(Labeling):
    """ """

    def __init__(self, context):
        super(PMLabeling, self).__init__(context)
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)

    def filter_manageable_labels(self, labels, modes=("view", "edit", )):
        """Give p_labels is like [[], []]."""
        # do not filter for Managers
        if not self.tool.isManager(realManagers=True):
            # filter depending on self._labels_cache
            cache = getattr(self.context, ITEM_LABELS_ACCESS_CACHE_ATTR)
            personal_labels = []
            global_labels = []
            user_groups = set(get_plone_groups_for_user())
            default_view_config_already_checked = default_edit_config_already_checked = False
            extra_expr_ctx = None
            view_expr_result_cache = {}
            edit_expr_result_cache = {}
            for label in labels[0]+labels[1]:
                # by_user labels are always editable
                if not label['by_user']:
                    # optimization if modes == ('view', ) and label not 'active'
                    # this is the case when displayed in faceted dashboard
                    if modes == ('view',) and not label['active']:
                        continue
                    # manage _labels_cache, if not in cache, use the config for
                    # 'default_for_all_labels'
                    cached = cache.get(label['label_id'])
                    is_using_default_config = False
                    if cached is None:
                        is_using_default_config = True
                        cached = cache["*"]
                    config = self.cfg.getLabelsConfig(["*"]) if \
                        is_using_default_config else \
                        self.cfg.getLabelsConfig([label['label_id']])
                    # first check for "view"
                    view_was_checked = False
                    if "view" in modes and \
                       label['active']:
                        if not is_using_default_config or \
                           default_view_config_already_checked is False:
                            # check "view_groups", in or not in depending on "view_groups_excluding"
                            if cached['view_groups'] and \
                               (
                                (config['view_groups_excluding'] == '0' and
                                 not user_groups.intersection(cached['view_groups'])) or
                                (config['view_groups_excluding'] == '1' and
                                 user_groups.intersection(cached['view_groups']))):
                                continue
                            # manage view_access, already computed, "False" or "True"
                            # "None" means needs to be computed on the fly
                            if cached['view_access'] is False:
                                continue
                            view_access_on = config['view_access_on'].strip()
                            if config['view_access_on_cache'] == '0' and view_access_on:
                                # will be done only on first use
                                if extra_expr_ctx is None:
                                    extra_expr_ctx = _base_extra_expr_ctx(
                                        self.context, {'item': self.context, })
                                # cache TAL expr result in case same TAL expr used for several configs
                                if view_access_on in view_expr_result_cache:
                                    if view_expr_result_cache[view_access_on] is False:
                                        continue
                                elif not _evaluateExpression(
                                        self.context,
                                        expression=view_access_on,
                                        extra_expr_ctx=extra_expr_ctx,
                                        raise_on_error=True):
                                    view_expr_result_cache[view_access_on] = False
                                    continue
                                view_expr_result_cache[view_access_on] = True
                            # mark view default config as working
                            if is_using_default_config:
                                default_view_config_already_checked = True
                            # mark "view" as checked, means "view" is in modes
                            # and check passed so we do not need to check "edit"
                            view_was_checked = True
                        else:
                            view_was_checked = True

                    # "edit" not having passed the "view" check
                    if "edit" in modes and not view_was_checked and \
                       (not is_using_default_config or
                            default_edit_config_already_checked is False):
                        # check "edit_groups", in or not in depending on "edit_groups_excluding"
                        if cached['edit_groups'] and \
                           (
                            (config['edit_groups_excluding'] == '0' and
                             not user_groups.intersection(cached['edit_groups'])) or
                            (config['edit_groups_excluding'] == '1' and
                             user_groups.intersection(cached['edit_groups']))):
                            continue
                        # manage edit_access_on, already computed, "False" or "True"
                        # "None" means needs to be computed on the fly
                        if cached['edit_access'] is False:
                            continue
                        edit_access_on = config['edit_access_on'].strip()
                        if config['edit_access_on_cache'] == '0' and edit_access_on:
                            # will be done only on first use
                            if extra_expr_ctx is None:
                                extra_expr_ctx = _base_extra_expr_ctx(
                                    self.context, {'item': self.context, })
                            # cache TAL expr result in case same TAL expr used for several configs
                            if edit_access_on in edit_expr_result_cache:
                                if edit_expr_result_cache[edit_access_on] is False:
                                    continue
                            elif not _evaluateExpression(
                                    self.context,
                                    expression=edit_access_on,
                                    extra_expr_ctx=extra_expr_ctx,
                                    raise_on_error=True):
                                edit_expr_result_cache[edit_access_on] = False
                                continue
                            edit_expr_result_cache[edit_access_on] = True
                        # mark edit default config as working
                        if is_using_default_config:
                            default_edit_config_already_checked = True
                    # if we are here, label may be kept, otherwise we would have
                    # encountered a "continue" here above
                    global_labels.append(label)
                else:
                    personal_labels.append(label)
            labels = [personal_labels, global_labels]
        return labels

    def get_searches_labels_cachekey(method, self):
        """Cachekey for self.get_searches_labels.
           When a DashboardCollection is added/modified/removed,
           the MeetingConfig modification date is updated."""
        return self.cfg.modified()

    @ram.cache(get_searches_labels_cachekey)
    def get_searches_labels(self):
        """Return list of labels used in searches of current MeetingConfig
           so searches using the "labels" index and for which showNumberOfItems
           is True. This is used to invalidate searches counter cache when
           changing a label but only when relevant."""
        res = []
        for collection in self.cfg.searches.searches_items.objectValues():
            if collection.showNumberOfItems is True and collection.enabled is True:
                parsed_query = parseFormquery(collection, collection.query)
                if "labels" in parsed_query:
                    # most queries use "query", except for "not" queries
                    res += parsed_query['labels'].get(
                        'query', parsed_query['labels'].get('not'))
        return res

    def pers_update(self, label_ids, activate):
        # invalidate collections counter cache and portlet_todo if any
        user_id = get_current_user_id(self.context.REQUEST)
        if set(['%s:%s' % (user_id, label_id) for label_id in label_ids]). \
                intersection(self.get_searches_labels()):
            invalidate_cachekey_volatile_for(
                'Products.PloneMeeting.MeetingItem.modified', get_again=True)
        return super(PMLabeling, self).pers_update(label_ids, activate)

    def update(self, label_ids):
        """Avoid labels not manageable to be removed, add it back to label_ids."""
        active_labels = self.active_labels()
        stored_label_ids = IAnnotations(self.context).get(FTW_LABELS_ANNOTATION_KEY, {}).keys()
        for active_label in active_labels:
            active_label['active'] = True
        # need a full label to filter it, returns pers and global labels
        active_label_ids = [label['label_id'] for label in active_labels
                            if not label['by_user']]
        # build a filter_manageable_labels compliant list of dicts
        new_labels = [
            {'label_id': label_id, 'active': True, 'by_user': False}
            for label_id in label_ids]
        editable_labels = self.filter_manageable_labels(
            [[], active_labels + new_labels], modes=('edit', ))[1]
        editable_label_ids = [label['label_id'] for label in editable_labels]
        # wipeout label_ids if not stored and not editable, this could mean
        # a label manipulated in the UI as this should not be possible
        label_ids = [label_id for label_id in label_ids
                     if label_id in stored_label_ids or label_id in editable_label_ids]
        viewable_labels = self.filter_manageable_labels(
            [[], active_labels], modes=('view', ))[1]
        viewable_label_ids = [label['label_id'] for label in viewable_labels]
        not_manageable_label_ids = set(active_label_ids).difference(
            editable_label_ids + label_ids)
        # ignore not viewable label ids
        viewable_not_manageable_label_ids = [
            not_mangeable_label_id for not_mangeable_label_id in not_manageable_label_ids
            if not_mangeable_label_id in viewable_label_ids]
        if viewable_not_manageable_label_ids:
            not_manageable_label_titles = [
                '"{0}"'.format(label['title']) for label in active_labels
                if label['label_id'] in viewable_not_manageable_label_ids]
            api.portal.show_message(
                _("You can not manage labels ${not_manageable_label_titles}!",
                  mapping={'not_manageable_label_titles': safe_unicode(
                    ', '.join(not_manageable_label_titles))}),
                type='warning',
                request=self.context.REQUEST)
        label_ids += list(not_manageable_label_ids)
        # check if need to update_local_roles, a relevant label has been (un)selected
        # check if one added or removed
        added_or_removed = tuple(set(label_ids).symmetric_difference(stored_label_ids))
        # this will add/remove relevant labels before eventual update_local_roles
        super(PMLabeling, self).update(label_ids)
        # first element of config is related to "*"
        config = self.cfg.getLabelsConfig(('*', ) + added_or_removed, data="update_local_roles")
        # we have:
        # a config other than the default specifying to update local roles
        # or we do not have "0" for every selected elements and
        # default config specify to update local roles
        if ("1" in config[1:]) or \
           (len(config)-1 != len(added_or_removed) and config[0] == "1"):
            self.context.update_local_roles(avoid_reindex=True)
        # invalidate collections counter cache and portlet_todo
        if set(active_label_ids).symmetric_difference(label_ids). \
                intersection(self.get_searches_labels()):
            invalidate_cachekey_volatile_for(
                'Products.PloneMeeting.MeetingItem.modified', get_again=True)



# -*- coding: utf-8 -*-
#
# GNU General Public License (GPL)
#

from AccessControl import Unauthorized
from imio.helpers.cache import invalidate_cachekey_volatile_for
from imio.helpers.security import fplog
from plone import api
from plone.z3cform.layout import wrap_form
from Products.CMFPlone.utils import safe_unicode
from Products.PloneMeeting.config import PMMessageFactory as _
from Products.PloneMeeting.interfaces import IRedirect
from Products.PloneMeeting.utils import _itemNumber_to_storedItemNumber
from Products.PloneMeeting.utils import notifyModifiedAndReindex
from Products.PloneMeeting.utils import validate_item_assembly_value
from z3c.form import button
from z3c.form import field
from z3c.form import form
from z3c.form.contentprovider import ContentProviders
from z3c.form.interfaces import IFieldsAndContentProvidersForm
from zope import interface
from zope import schema
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.component.hooks import getSite
from zope.contentprovider.provider import ContentProviderBase
from zope.i18n import translate
from zope.interface import implements


USING_ABSENTS_OR_EXCUSED_MSGID = u"Enter the item attendees to be applied. By default, the value of the field is " \
    "what is defined on the meeting. If you do not change this value, nothing will be applied on the item. If you " \
    "already edited this field before and you want to fallback to meeting value, remove the entire value."
USING_ONLY_ASSEMBLY_MSGID = u"Enter the item assembly to be applied. By default, the value of the field is what is " \
    "defined on the meeting. If you do not change this value, nothing will be applied on the item. If you already " \
    "edited this field before and you want to fallback to meeting value, remove the entire value.  You may add [[ ]] " \
    "around absent people (like [[Mister Sample Peter]])."


def item_assembly_default():
    """
      Returns the itemAssembly of the item.
      As from zope.schema._bootstrapinterfaces import
      IContextAwareDefaultFactory does not seem to work,
      we have to get current context manually...
    """
    context = getSite().REQUEST['PUBLISHED'].context
    itemAssembly = context.getItemAssembly(for_display=False)
    # need to strip because extractData strips
    return safe_unicode(itemAssembly).strip()


def item_excused_default():
    """
      Returns the itemAssemblyExcused of the item.
      As from zope.schema._bootstrapinterfaces import
      IContextAwareDefaultFactory does not seem to work,
      we have to get current context manually...
    """
    context = getSite().REQUEST['PUBLISHED'].context
    itemAssemblyExcused = context.getItemAssemblyExcused(for_display=False)
    # need to strip because extractData strips
    return safe_unicode(itemAssemblyExcused).strip()


def item_absents_default():
    """
      Returns the itemAssemblyAbsents of the item.
      As from zope.schema._bootstrapinterfaces import
      IContextAwareDefaultFactory does not seem to work,
      we have to get current context manually...
    """
    context = getSite().REQUEST['PUBLISHED'].context
    itemAssemblyAbsents = context.getItemAssemblyAbsents(for_display=False)
    # need to strip because extractData strips
    return safe_unicode(itemAssemblyAbsents).strip()


def item_guests_default():
    """
      Returns the itemAssemblyGuests of the item.
      As from zope.schema._bootstrapinterfaces import
      IContextAwareDefaultFactory does not seem to work,
      we have to get current context manually...
    """
    context = getSite().REQUEST['PUBLISHED'].context
    itemAssemblyGuests = context.getItemAssemblyGuests(for_display=False)
    # need to strip because extractData strips
    return safe_unicode(itemAssemblyGuests).strip()


def validate_apply_until_item_number(value):
    '''This method does validate the 'apply_until_item_number' field.
       It will check that given number is a number among existing item numbers
       using a correct format.'''
    try:
        _itemNumber_to_storedItemNumber(value)
    except ValueError:
        raise interface.Invalid(_(u'Please provide a valid number.'))
    return True


def validate_item_assembly(value):
    '''This method does validate the 'item_assembly' field.
       It will check that [[]] are correct.'''
    if not validate_item_assembly_value(value):
        raise interface.Invalid(_(u'Please check that opening "[[" have corresponding closing "]]".'))
    return True


class IManageItemAssembly(interface.Interface):
    item_assembly = schema.Text(
        title=_(u"Item assembly to apply"),
        description=_(USING_ONLY_ASSEMBLY_MSGID),
        defaultFactory=item_assembly_default,
        required=False,
        constraint=validate_item_assembly)
    item_excused = schema.Text(
        title=_(u"Item excused to apply"),
        description=_(u"Enter the item excused to be applied. "
                      u"By default, the value of the field is what is defined on the meeting. "
                      u"If you do not change this value, nothing will be applied on the item. "
                      u"If you already edited this field before and you want to fallback to meeting value, "
                      u"remove the entire value."),
        defaultFactory=item_excused_default,
        required=False,)
    item_absents = schema.Text(
        title=_(u"Item absents to apply"),
        description=_(u"Enter the item absents to be applied. "
                      u"By default, the value of the field is what is defined on the meeting. "
                      u"If you do not change this value, nothing will be applied on the item. "
                      u"If you already edited this field before and you want to fallback to meeting value, "
                      u"remove the entire value."),
        defaultFactory=item_absents_default,
        required=False,)
    item_guests = schema.Text(
        title=_(u"Item guests to apply"),
        description=_(u"Enter the item guests to be applied. "
                      u"By default, the value of the field is what is defined on the meeting. "
                      u"If you do not change this value, nothing will be applied on the item. "
                      u"If you already edited this field before and you want to fallback to meeting value, "
                      u"remove the entire value."),
        defaultFactory=item_guests_default,
        required=False,)

    apply_until_item_number = schema.TextLine(
        title=_(u"Apply until item number"),
        description=_(u"If you specify a number, the values entered here above will be applied from current "
                      u"item to the item number entered. Leave empty to only apply for current item."),
        required=False,
        constraint=validate_apply_until_item_number,)


class DisplayAssemblyFromMeetingProvider(ContentProviderBase):
    """
      This ContentProvider will just display
      the assembly defined on the linked meeting.
    """
    template = \
        ViewPageTemplateFile('templates/display_assembly_from_meeting.pt')

    def __init__(self, context, request, view):
        super(DisplayAssemblyFromMeetingProvider, self).__init__(context,
                                                                 request,
                                                                 view)
        self.__parent__ = view

    def get_meeting_assembly(self):
        """
          Return Meeting.assembly
        """
        meeting = self.context.getMeeting()
        nothing_defined_msg = translate('nothing_defined_on_meeting',
                                        domain='PloneMeeting',
                                        context=self.request)
        return meeting.get_assembly(
            for_display=True, striked=False, mark_empty_tags=True) or \
            u'<p class="discreet">{0}</p>'.format(nothing_defined_msg)

    def get_msgid_assembly_or_attendees(self):
        """
          Return the msgid to translate, either 'Assembly' or
          'Attendees' defined on the meeting.
        """
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self.context)
        usedMeetingAttributes = cfg.getUsedMeetingAttributes()
        if 'assembly_excused' in usedMeetingAttributes or \
           'assembly_absents' in usedMeetingAttributes:
            return 'display_meeting_attendees_legend'
        else:
            return 'display_meeting_assembly_legend'

    def render(self):
        # field may not be used and form is used to manage guests
        if self.context.is_assembly_field_used('itemAssembly'):
            return self.template()
        else:
            return ''


class DisplayExcusedFromMeetingProvider(ContentProviderBase):
    """
      This ContentProvider will just display
      the excused defined on the linked meeting.
    """
    template = \
        ViewPageTemplateFile('templates/display_excused_from_meeting.pt')

    def __init__(self, context, request, view):
        super(DisplayExcusedFromMeetingProvider, self).__init__(context,
                                                                request,
                                                                view)
        self.__parent__ = view

    def get_assembly_excused(self):
        """
          Return Meeting.assembly_excused
        """
        meeting = self.context.getMeeting()
        nothing_defined_msg = translate('nothing_defined_on_meeting',
                                        domain='PloneMeeting',
                                        context=self.request)
        return meeting.get_assembly_excused(
            for_display=True, striked=False, mark_empty_tags=True) or \
            u'<p class="discreet">{0}</p>'.format(nothing_defined_msg)

    def render(self):
        if self.context.is_assembly_field_used('itemAssemblyExcused'):
            return self.template()
        else:
            return ''


class DisplayAbsentsFromMeetingProvider(ContentProviderBase):
    """
      This ContentProvider will just display
      the absents defined on the linked meeting.
    """
    template = \
        ViewPageTemplateFile('templates/display_absents_from_meeting.pt')

    def __init__(self, context, request, view):
        super(DisplayAbsentsFromMeetingProvider, self).__init__(
            context, request, view)
        self.__parent__ = view

    def get_assembly_absents(self):
        """
          Return Meeting.assembly_absents
        """
        meeting = self.context.getMeeting()
        nothing_defined_msg = translate('nothing_defined_on_meeting',
                                        domain='PloneMeeting',
                                        context=self.request)
        return meeting.get_assembly_absents(
            for_display=True, striked=False, mark_empty_tags=True) or \
            u'<p class="discreet">{0}</p>'.format(nothing_defined_msg)

    def render(self):
        if self.context.is_assembly_field_used('itemAssemblyAbsents'):
            return self.template()
        else:
            return ''


def _itemsToUpdate(from_item_number, until_item_number, meeting):
    """
      Return items we want to update regarding the number
      defined in apply_until_item_number
    """
    catalog = api.portal.get_tool('portal_catalog')
    if not until_item_number:
        until_item_number = from_item_number
    brains = catalog.unrestrictedSearchResults(
        meeting_uid=meeting.UID(),
        getItemNumber={'query': (from_item_number,
                                 until_item_number),
                       'range': 'minmax'},
        sort_on='getItemNumber')
    return [brain.getObject() for brain in brains]


class ManageItemAssemblyForm(form.Form):
    """
      This form will help MeetingManagers manage itemAssembly
      by being able to redefine it on a single
      item without having to use the edit form and to apply
      redefined value until the item number he wants.
    """
    implements(IFieldsAndContentProvidersForm)

    fields = field.Fields(IManageItemAssembly)
    ignoreContext = True  # don't use context to get widget data

    contentProviders = ContentProviders()
    contentProviders['assembly'] = DisplayAssemblyFromMeetingProvider
    contentProviders['assembly'].position = 0
    contentProviders['excused'] = DisplayExcusedFromMeetingProvider
    contentProviders['excused'].position = 2
    contentProviders['absents'] = DisplayAbsentsFromMeetingProvider
    contentProviders['absents'].position = 2

    label = _(u"Manage item assembly")
    description = u''
    _finished = False

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.label = translate('Manage item assembly',
                               domain='PloneMeeting',
                               context=self.request)

    @button.buttonAndHandler(_('Apply'), name='apply_item_assembly')
    def handleApplyItemAssembly(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        # do adapt item assembly
        self.item_assembly = data.get('item_assembly')
        self.item_excused = data.get('item_excused')
        self.item_absents = data.get('item_absents')
        self.item_guests = data.get('item_guests')
        # we receive '5' or '5.2' but we want 500 or 502
        self.apply_until_item_number = \
            _itemNumber_to_storedItemNumber(
                data.get('apply_until_item_number') or u'0'
            )
        self.meeting = self.context.getMeeting()
        self._doApplyItemAssembly()

    @button.buttonAndHandler(_('Cancel'), name='cancel')
    def handleCancel(self, action):
        self._finished = True

    def _check_auth(self):
        """Raise Unauthorized if current user can not manage itemAssembly."""
        if not self.context.mayQuickEditItemAssembly():
            raise Unauthorized

    def update(self):
        """ """
        self._check_auth()
        super(ManageItemAssemblyForm, self).update()
        # after calling parent's update, self.actions are available
        self.actions.get('cancel').addClass('standalone')

    def updateWidgets(self):
        # XXX manipulate self.fields BEFORE doing form.Form.updateWidgets
        # show only relevant fields
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self.context)
        usedMeetingAttributes = cfg.getUsedMeetingAttributes()
        self.fields['item_assembly'].mode = 'hidden'
        self.fields['item_excused'].mode = 'hidden'
        self.fields['item_absents'].mode = 'hidden'
        self.fields['item_guests'].mode = 'hidden'
        changeItemAssemblyTitleAndDescr = False
        # this form is also used to edit only 'guests' when using attendees
        # manage also when switching from assembly to attendees
        # "assembly" field may be disabled but assembly used on meeting
        if 'assembly' in usedMeetingAttributes or \
                self.context.getItemAssembly():
            self.fields['item_assembly'].mode = 'input'
        if 'assembly_excused' in usedMeetingAttributes or \
                self.context.getItemAssemblyExcused():
            changeItemAssemblyTitleAndDescr = True
            self.fields['item_excused'].mode = 'input'
        if 'assembly_absents' in usedMeetingAttributes or \
                self.context.getItemAssemblyAbsents():
            changeItemAssemblyTitleAndDescr = True
            self.fields['item_absents'].mode = 'input'
        if 'assembly_guests' in usedMeetingAttributes or \
                self.context.getItemAssemblyGuests():
            changeItemAssemblyTitleAndDescr = True
            self.fields['item_guests'].mode = 'input'
        if changeItemAssemblyTitleAndDescr:
            self.fields['item_assembly'].field.title = \
                _('Item attendees to apply')
            self.fields['item_assembly'].field.description = \
                _(USING_ABSENTS_OR_EXCUSED_MSGID)
        else:
            self.fields['item_assembly'].field.title = \
                _('Item assembly to apply')
            self.fields['item_assembly'].field.description = \
                _(USING_ONLY_ASSEMBLY_MSGID)
        form.Form.updateWidgets(self)

    def render(self):
        if self._finished:
            IRedirect(self.request).redirect(self.context.absolute_url())
            return ""
        return super(ManageItemAssemblyForm, self).render()

    def _doApplyItemAssembly(self):
        """
          The method actually do the job, set the itemAssembly on self.context
          and following items if defined
        """
        self._check_auth()
        # only update if default proposed value was changed
        item_assembly_def = item_assembly_default()
        item_excused_def = item_excused_default()
        item_absents_def = item_absents_default()
        item_guests_def = item_guests_default()
        from_item_number = self.context.getItemNumber(relativeTo='meeting')
        until_item_number = self.apply_until_item_number
        items_to_update = _itemsToUpdate(
            from_item_number=from_item_number,
            until_item_number=until_item_number,
            meeting=self.meeting)
        for itemToUpdate in items_to_update:
            # only update if we changed default value
            if self.item_assembly != item_assembly_def:
                itemToUpdate.setItemAssembly(self.item_assembly)
            if self.item_excused != item_excused_def:
                itemToUpdate.setItemAssemblyExcused(self.item_excused)
            if self.item_absents != item_absents_def:
                itemToUpdate.setItemAssemblyAbsents(self.item_absents)
            if self.item_guests != item_guests_def:
                itemToUpdate.setItemAssemblyGuests(self.item_guests)
            notifyModifiedAndReindex(itemToUpdate)

        # invalidate assembly async load on item
        invalidate_cachekey_volatile_for(
            'Products.PloneMeeting.browser.async.AsyncLoadItemAssemblyAndSignaturesRawFields',
            get_again=True)

        first_item_number = items_to_update[0].getItemNumber(for_display=True)
        last_item_number = items_to_update[-1].getItemNumber(for_display=True)
        extras = 'item={0} from_item_number={1} until_item_number={2}'.format(
            repr(self.context), first_item_number, last_item_number)
        fplog('manage_item_assembly', extras=extras)
        api.portal.show_message(_("Item assemblies have been updated."), request=self.request)
        self._finished = True


ManageItemAssemblyFormWrapper = wrap_form(ManageItemAssemblyForm)



# -*- coding: utf-8 -*-
#
# File: indexes.py
#
# GNU General Public License (GPL)
#

from collective.contact.core.content.organization import IOrganization
from collective.iconifiedcategory.indexes import content_category_uid
from datetime import datetime
from imio.annex.content.annex import IAnnex
from imio.helpers import EMPTY_STRING
from imio.helpers.content import _contained_objects
from imio.helpers.content import base_getattr
from imio.helpers.content import safe_delattr
from imio.history.utils import getLastWFAction
from imio.pyutils.utils import safe_encode
from OFS.interfaces import IItem
from plone import api
from plone.indexer import indexer
from Products.PloneMeeting.config import ALL_ADVISERS_GROUP_VALUE
from Products.PloneMeeting.config import HIDDEN_DURING_REDACTION_ADVICE_VALUE
from Products.PloneMeeting.config import ITEM_INITIATOR_INDEX_PATTERN
from Products.PloneMeeting.config import ITEM_NO_PREFERRED_MEETING_VALUE
from Products.PloneMeeting.config import NOT_GIVEN_ADVICE_VALUE
from Products.PloneMeeting.config import REINDEX_NEEDED_MARKER
from Products.PloneMeeting.content.meeting import IMeeting
from Products.PloneMeeting.interfaces import IMeetingContent
from Products.PloneMeeting.interfaces import IMeetingItem
from Products.PloneMeeting.utils import get_annexes
from Products.PloneMeeting.utils import get_datagridfield_column_value
from Products.PluginIndexes.common.UnIndex import _marker


REAL_ORG_UID_PATTERN = 'real_org_uid__{0}'
DELAYAWARE_ROW_ID_PATTERN = 'delay_row_id__{0}'


@indexer(IItem)
def getConfigId(obj):
    """
      Indexes the MeetingConfig id.
    """
    tool = api.portal.get_tool('portal_plonemeeting')
    cfg = tool.getMeetingConfig(obj)
    return cfg and cfg.getId() or EMPTY_STRING


@indexer(IMeetingItem)
def getTakenOverBy(obj):
    """
      Indexes the takenOverBy with a special value when empty,
      in faceted filters, query on empty value is ignored...
    """
    takenOverBy = obj.getTakenOverBy()
    if not takenOverBy:
        takenOverBy = EMPTY_STRING
    return takenOverBy


@indexer(IMeeting)
def sortable_title(obj):
    """
      Indexes the sortable_title of meeting based on meeting.date
    """
    return obj.date.strftime('%Y%m%d%H%M')


@indexer(IMeeting)
def meeting_date(obj):
    """
      Indexes the meeting.date
    """
    return obj.date


@indexer(IMeetingItem)
def title_or_id(obj):
    """
      Indexes the title_or_id, used in the referencebrowser popup
    """
    return obj.title_or_id(withTypeName=False)


@indexer(IMeetingContent)
def previous_review_state(obj):
    """
      Indexes the previous review_state, aka the review_state before current review_state
    """
    previous_action = None
    event = getLastWFAction(obj, transition='before_last')
    if event:
        previous_action = event['review_state']
    return previous_action or _marker


@indexer(IMeetingContent)
def pm_technical_index(obj):
    """
    This index is made to hold various technical informations:
    - "reindex_needed" is used to know if a reindex must occurs during night tasks;
    - "item_initiator_..." will hold MeetingItem.itemInititors values to ease
      removing a held_position that is not used;
    - "item_linked_items_reorder_needed" is used to know if reordering linked items
      is needed on some items, this is the case when items that were already linked
      togheter are presented to a meeting as linked items are sorted by meeting date;
    - ...
    """
    res = []
    if base_getattr(obj, REINDEX_NEEDED_MARKER, False):
        res.append(REINDEX_NEEDED_MARKER)
    if obj.__class__.__name__ == "MeetingItem":
        for hp_uid in obj.getItemInitiator():
            res.append(ITEM_INITIATOR_INDEX_PATTERN.format(hp_uid))
    return res or _marker


@indexer(IMeetingItem)
def Description(obj):
    """
      Make sure to use 'text/plain' version of description field as it is normally
      a TextField and that we store HTML data into it for MeetingItem
    """
    return obj.Description(mimetype='text/plain')


@indexer(IMeetingItem)
def getCopyGroups(obj):
    """
      Compute copyGroups and restrictedCopyGroups to take auto copyGroups into account.
      restrictedCopyGroups are prefixed by restricted__ so it can be displayed differently
      in dashboard filter and column.
    """
    return obj.getAllCopyGroups(auto_real_plone_group_ids=True) + \
        obj.getAllRestrictedCopyGroups(auto_real_plone_group_ids=True) or EMPTY_STRING


@indexer(IMeetingItem)
def reviewProcessInfo(obj):
    """
      Compute a reviewProcessInfo, this concatenate the group managing item
      and the item review_state so it can be queryable in the catalog.
    """
    item_state = obj.query_state()
    return '%s__reviewprocess__%s' % (
        obj.adapted()._getGroupManagingItem(item_state, theObject=False), item_state)


@indexer(IMeetingItem)
def meeting_uid(obj):
    """
      Store the linked meeting UID.
    """
    # we use same 'None' value as for getPreferredMeeting so we may use the same
    # vocabulary in the meeting date/preferred meeting date faceted filters
    res = ITEM_NO_PREFERRED_MEETING_VALUE
    meeting_uid = obj.getMeeting(only_uid=True)
    if meeting_uid:
        res = meeting_uid
    return res


@indexer(IMeetingItem)
def item_meeting_date(obj):
    """
      Store the linked meeting date.
    """
    res = []
    meeting = obj.getMeeting()
    if meeting:
        res = meeting.date
    else:
        # for sorting it is necessary to have a date
        res = datetime(1950, 1, 1)
    return res


@indexer(IMeetingItem)
def getGroupsInCharge(obj):
    """
      Indexes the groupsInCharge attribute but do not include auto groups
      as it is stored on obj
    """
    return obj.getGroupsInCharge(includeAuto=False) or EMPTY_STRING


@indexer(IMeetingItem)
def preferred_meeting_uid(obj):
    """
      Store the preferredMeeting.
    """
    preferredMeeting = obj.getPreferredMeeting()
    return preferredMeeting


@indexer(IMeetingItem)
def preferred_meeting_date(obj):
    """
      Store the preferredMeeting date.
    """
    res = []
    preferredMeeting = obj.getPreferredMeeting(theObject=True)
    if preferredMeeting:
        res = preferredMeeting.date
    else:
        res = datetime(1950, 1, 1)
    return res or _marker


@indexer(IItem)
def sentToInfos(obj):
    """
      Index other meetingConfigs the item will be/has been cloned to.
      We append :
      - __clonable_to to a meetingConfig id the item is clonable to;
      - __cloned_to to a meetingConfig id the item is cloned to.
      An item that does not have to be send to another meetingConfig
      will receive the 'not_to_be_cloned_to' value so we can filter it out.
    """
    res = []
    clonableTo = obj.getOtherMeetingConfigsClonableTo()
    clonableToEmergency = obj.getOtherMeetingConfigsClonableToEmergency()
    clonedTo = obj._getOtherMeetingConfigsImAmClonedIn()
    for cfgId in clonableTo:
        if cfgId not in clonedTo:
            term_suffix = '__clonable_to'
            if cfgId in clonableToEmergency:
                term_suffix = '__clonable_to_emergency'
            res.append(cfgId + term_suffix)
    for cfgId in clonedTo:
        term_suffix = '__cloned_to'
        if cfgId in clonableToEmergency:
            term_suffix = '__cloned_to_emergency'
        res.append(cfgId + term_suffix)
    if not clonableTo and not clonedTo:
        res.append('not_to_be_cloned_to')
    return res or _marker


def SearchableText(obj):
    """
      Contained annex title and scan_id are indexed in the SearchableText.
    """
    res = []
    res.append(obj.SearchableText())
    for annex in get_annexes(obj):
        res.append(annex.Title())
        scan_id = getattr(annex, "scan_id", None)
        if scan_id:
            # we need utf-8, when edited thru the UI, scan_id is stored as unicode
            res.append(safe_encode(scan_id))
    res = ' '.join(res)
    return res or _marker


@indexer(IMeetingItem)
def SearchableText_item(obj):
    # remove reindex needed marker
    safe_delattr(obj, REINDEX_NEEDED_MARKER)
    return SearchableText(obj)


@indexer(IAnnex)
def SearchableText_annex(obj):
    return _marker


@indexer(IMeetingItem)
def send_to_authority(obj):
    """
      Index the MeetingItem.sendToAuthority to be searchable in a faceted navigation.
    """
    if obj.getSendToAuthority():
        return '1'
    else:
        return '0'


@indexer(IMeetingItem)
def item_is_signed(obj):
    """
      Index the getItemIsSigned but in a faceted boolean compatible way.
    """
    if obj.getItemIsSigned():
        return '1'
    return '0'


@indexer(IMeetingItem)
def to_discuss(obj):
    """
      Index the getToDiscuss but in a faceted boolean compatible way.
    """
    if obj.getToDiscuss():
        return '1'
    return '0'


@indexer(IMeetingItem)
def annexes_index(obj):
    """
      Unique index with data relative to stored annexes :
      - to_print :
        - 'not_to_print' if contains annexes not to print;
        - 'to_print' otherwise;
      - confidential :
        - 'not_confidential' if contains not confidential annexes;
        - 'confidential' otherwise;
      - publishable :
        - 'not_publishable' if contains not publishable annexes;
        - 'publishable' otherwise;
      - to_sign/signed :
        - 'not_to_sign' if contains not to sign annexes;
        - 'to_sign' if contains to_sign but not signed annexes;
        - 'signed' if contains signed annexes.
    """
    res = []
    # use objectValues because with events order, an annex
    # could be added but still not registered in the categorized_elements dict
    for annex in get_annexes(obj):
        # to_print
        if annex.to_print:
            res.append('to_print')
        else:
            res.append('not_to_print')
        # confidential
        if annex.confidential:
            res.append('confidential')
        else:
            res.append('not_confidential')
        # publishable
        if annex.publishable:
            res.append('publishable')
        else:
            res.append('not_publishable')
        # to_sign/signed
        if annex.to_sign:
            if annex.signed:
                res.append('signed')
            else:
                res.append('to_sign')
        else:
            res.append('not_to_sign')
    # remove duplicates
    return list(set(res))


@indexer(IItem)
def templateUsingGroups(obj):
    """
      Index used to build the item templates tree.
      If not attribute 'templateUsingGroups' (so not a MeetingItem)
      or a MeetingItem with no selected templateUsingGroups.
      In the query, we will query '__nothing_selected__' + organizations the current
      user is creator for.
    """
    if obj.meta_type == 'MeetingItem':
        templateUsingGroups = obj.getTemplateUsingGroups()
        return templateUsingGroups and templateUsingGroups or ('__nothing_selected__', )
    elif obj.portal_type == 'Folder':
        return ('__folder_in_itemtemplates__', )
    return _marker


def _to_coded_adviser_index(obj, org_uid, advice_infos):
    """Build an 'index' version of the adviser state so it is searchable and so on."""
    def _computeSuffixFor(org_uid, advice_infos, advice_type):
        '''
          Compute the suffix that will be appended depending on advice state.
        '''
        suffixes = []
        # still not given but still giveable?  Not giveable?  Delay exceeded? Asked again?
        if advice_type in (NOT_GIVEN_ADVICE_VALUE,
                           'asked_again',
                           HIDDEN_DURING_REDACTION_ADVICE_VALUE):
            if obj._adviceDelayIsTimedOut(org_uid):
                # delay is exceeded, advice was not given
                suffixes.append('_advice_delay_exceeded')
            else:
                # does the relevant organization may add the advice in current item state?
                if advice_infos['advice_addable']:
                    suffixes.append('_advice_not_given')
                elif advice_infos['advice_editable']:
                    if advice_type == 'asked_again':
                        suffixes.append('_advice_asked_again')
                    elif advice_type == HIDDEN_DURING_REDACTION_ADVICE_VALUE:
                        suffixes.append('_advice_{0}'.format(HIDDEN_DURING_REDACTION_ADVICE_VALUE))
                else:
                    suffixes.append('_advice_not_giveable')

        if advice_infos['type'] != NOT_GIVEN_ADVICE_VALUE:
            # if advice was given, is it still editable or not?
            # we return the current advice review_state
            # by default, a still editable advice is 'advice_under_edit'
            # and a no more editable advice is 'advice_given'
            advice_obj = getattr(obj, advice_infos['advice_id'])
            suffixes.append('_%s' % advice_obj.query_state())
        return suffixes

    res = []
    isDelayAware = advice_infos['delay'] and True or False
    # compute suffixes
    # we compute the 'advice_type' to take into account 'hidden_during_redaction'
    advice_type = obj._shownAdviceTypeFor(advice_infos)
    suffixes = _computeSuffixFor(org_uid, advice_infos, advice_type)
    # we also index the 'real_org_uid_' so we can query who we asked
    # advice to, without passing the advice state
    for suffix in suffixes:
        if isDelayAware:
            res.append('delay__' + org_uid + suffix)
            res.append('delay__' + org_uid + suffix + '__' + advice_type)
            # 'delay_row_id_'
            delay_row_id = DELAYAWARE_ROW_ID_PATTERN.format(advice_infos['row_id'])
            res.append(delay_row_id)
            # 'delay_row_id_' with suffixed advice_type
            res.append(delay_row_id + '__' + advice_type)
        else:
            res.append(org_uid + suffix)
            res.append(org_uid + suffix + '__' + advice_type)
            # 'real_org_uid_'
            real_org_uid = REAL_ORG_UID_PATTERN.format(org_uid)
            res.append(real_org_uid)
            # 'real_org_uid_' with suffixed advice_type
            res.append(real_org_uid + '__' + advice_type)
    # userids
    tmp_res = list(res)
    # still need to check if key userids is there for older adviceIndex without userids
    if advice_infos.get('userids', None):
        for userid in advice_infos['userids']:
            for elt in tmp_res:
                res.append('{0}__userid__{1}'.format(elt, userid))
    else:
        # add a special value when advice only asked to entire advisers group
        for elt in tmp_res:
            res.append('{0}__userid__{1}'.format(elt, ALL_ADVISERS_GROUP_VALUE))

    # advice_type
    if advice_type not in res:
        res.append(advice_type)
    return res


@indexer(IMeetingItem)
def indexAdvisers(obj):
    """
      Build the index specifying advices to give.
      Values are different if it is a delay-aware or not advice :
      Delay-aware advice is like "delay__[developers_UID]_advice_not_given":
      - delay__ specifies that it is a delay-aware advice;
      - [developers_UID] is the UID of the organization the advice is asked to;
      Non delay-aware advice is like "[developers_UID]_advice_not_given".
      In both cases (delay-aware or not), we have a suffix :
        - '_advice_not_giveable' for advice not given and not giveable;
        - '_advice_not_given' for advice not given/asked again but giveable;
        - '_advice_delay_exceeded' for delay-aware advice not given but
           no more giveable because of delay exceeded;
    """
    if not hasattr(obj, 'adviceIndex'):
        return _marker

    res = []
    for org_uid, advice_infos in obj.adviceIndex.iteritems():
        adviceHolder = obj
        # use original advice data if advice is inherited
        if obj.adviceIsInherited(org_uid):
            advice_infos = obj.getInheritedAdviceInfo(org_uid)
            # when an advice is inherited from an already inherited advice,
            # when pasting new item, the predecessor is still not set so no adviceHolder
            # is available, in this case, we continue as item will be reindexed after
            if not advice_infos:
                continue
            adviceHolder = advice_infos['adviceHolder']
        res += _to_coded_adviser_index(adviceHolder, org_uid, advice_infos)
    # remove double entry, it could be the case for the 'advice_type' alone
    res = list(set(res))
    res.sort()
    # store something when no advices. Query with 'not' in ZCatalog>=3 will retrieve it.
    if not res:
        res.append('_')
    return res


@indexer(IOrganization)
def get_full_title(obj):
    '''By default we hide the "My organization" level, but not in the indexed
       value as it is used in the contact widget.'''
    return obj.get_full_title(force_separator=True)


@indexer(IMeetingItem)
def contained_uids_item(obj):
    """
      Indexes the UID of every contained elements.
    """
    return [contained.UID() for contained in _contained_objects(obj)] or _marker


@indexer(IMeeting)
def contained_uids_meeting(obj):
    """
      Indexes the UID of every contained elements.
    """
    return [contained.UID() for contained in _contained_objects(obj)] or _marker


@indexer(IMeetingItem)
def content_category_uid_item(obj):
    """
      Indexes the content_category of every contained elements.
    """
    return content_category_uid(obj)


@indexer(IMeetingItem)
def committees_index_item(obj):
    """
      Indexes the committees of an item into "committees_index" index.
    """
    return obj.getCommittees() or EMPTY_STRING


@indexer(IMeeting)
def committees_index_meeting(obj):
    """
      Indexes the committees of a meeting into "committees_index" index.
    """
    return get_datagridfield_column_value(obj.committees, "row_id") or EMPTY_STRING


@indexer(IMeeting)
def getCategory(obj):
    """
      Indexes the category under name "getCategory" for Meetings.
    """
    # can not index "None"
    return obj.category or _marker



# -*- coding: utf-8 -*-

from ftw.labels.interfaces import ILabelSupport
from ftw.labels.labeling import ILabeling
from persistent.mapping import PersistentMapping
from Products.PloneMeeting.migrations import logger
from Products.PloneMeeting.migrations.migrate_to_4104 import Migrate_To_4104
from Products.ZCatalog.ProgressHandler import ZLogHandler
from zope.annotation.interfaces import IAnnotations


class Migrate_To_4105(Migrate_To_4104):

    def _removeBrokenAnnexes(self):
        """ """
        logger.info("Remove broken annexes, annexes uploaded withtout a content_category...")
        brains = self.catalog(portal_type=['annex', 'annexDecision'])
        i = 0
        idxs = ['modified', 'ModificationDate', 'Date']
        for brain in brains:
            if not brain.content_category_uid:
                annex = brain.getObject()
                logger.info('In _removeBrokenAnnexes, removed %s' % brain.getPath())
                # make sure parent is not modified
                parent = annex.aq_parent
                parent_modified = parent.modified()
                parent.manage_delObjects(ids=[annex.getId()])
                parent.setModificationDate(parent_modified)
                parent.reindexObject(idxs=idxs)
                i += 1
        if i:
            self.warn(logger, 'In _removeBrokenAnnexes, removed %s annexe(s)' % i)
        logger.info('Done.')

    def _uncatalogWrongBrains(self):
        """Probably because before we reindexed parent upon annex add/edit/delete,
           some wrong paths are stored in catalog, these paths ends with '/' and does not have
           a correct UID."""
        logger.info("Uncataloging wrong brains...")
        i = 0
        for path in self.catalog._catalog.uids.keys():
            if path.endswith('/'):
                rid = self.catalog._catalog.uids[path]
                metadata = self.catalog._catalog.getMetadataForRID(rid)
                if metadata['UID'] is None:
                    logger.info('In _uncatalogWrongBrains, uncataloged %s' % path)
                    self.catalog._catalog.uncatalogObject(path)
                    # fix UUIDIndex if correct path exists (without ending "/")
                    correct_rid = self.catalog._catalog.uids.get(path[:-1], None)
                    if correct_rid:
                        try:
                            obj = self.catalog.getobject(correct_rid)
                            index = self.catalog._catalog.getIndex('UID')
                            index.unindex_object(correct_rid)
                            index.index_object(correct_rid, obj)
                        except AttributeError:
                            self.warn(
                                logger,
                                'In _uncatalogWrongBrains, could not get correct_rid %s at %s'
                                % (correct_rid, path[:-1]))
                    i += 1
        if i:
            self.warn(logger, 'In _uncatalogWrongBrains, uncataloged %s path(s)' % i)
        logger.info('Done.')
        return i or -1

    def _cleanFTWLabels(self):
        """This fix partial migrations of stored ftw.labels:labeling that are
           still PersistentList and not PersistentMapping."""
        logger.info("Cleaning ftw.labels wrong annotations...")
        brains = self.catalog(object_provides=ILabelSupport.__identifier__)
        pghandler = ZLogHandler(steps=100)
        pghandler.init('Cleaning ftw.labels wrong annotations...', len(brains))
        i = 0
        cleaned = 0
        for brain in brains:
            i += 1
            pghandler.report(i)
            obj = brain.getObject()
            annotations = IAnnotations(obj)
            if 'ftw.labels:labeling' in annotations and \
               not isinstance(annotations['ftw.labels:labeling'], PersistentMapping):
                if annotations['ftw.labels:labeling']:
                    labeling = ILabeling(obj)
                    old_values = [label for label in labeling.storage]
                    del annotations['ftw.labels:labeling']
                    labeling._storage = None
                    labeling.update(old_values)
                    logger.info('In _cleanFTWLabels, cleaned %s' % brain.getPath())
                else:
                    del annotations['ftw.labels:labeling']
                obj.reindexObject(idxs=['labels'])
                cleaned += 1
        if cleaned:
            self.warn(logger, 'In _cleanFTWLabels, cleaned %s element(s)' % cleaned)
        pghandler.finish()
        logger.info('Done.')

    def run(self, from_migration_to_41=False):
        logger.info('Migrating to PloneMeeting 4105...')
        # need to uncatalog wrong brains as long as there are wrong brains to uncatalog...
        uncatalogued = 0
        while uncatalogued != -1:
            uncatalogued = self._uncatalogWrongBrains()
        self._removeBrokenAnnexes()
        self._cleanFTWLabels()
        # from Migrate_To_4104
        self._moveMCParameterToWFA()
        self._addItemNonAttendeesAttributeToMeetings()
        # reapply MeetingItem.xml before reloadMeetingConfigs
        # that will remove the 'Duplicate and keep link' action
        self.ps.runImportStepFromProfile('profile-Products.PloneMeeting:default', 'typeinfo')
        # reload MeetingConfigs as some actions changed
        self.reloadMeetingConfigs()
        self.upgradeAll(omit=['Products.PloneMeeting:default',
                              self.profile_name.replace('profile-', '')])


def migrate(context):
    '''This migration function will:

       1) Clean wrong paths in catalog (ending with '/');
       2) Remove broken annexes with no content_category defined due to quickupload ConflictError management;
       3) Clean ftw.labels empty annotations, this make sure to not have any PersistentList stored;
       5) From Migrate_To_4104, move MeetingConfig.meetingManagerMayCorrectClosedMeeting to a workflowAdaptation;
       6) From Migrate_To_4104, add new attribute 'itemNonAttendees' to every meetings;
       7) Update portal_types to remove action 'Duplicate and keep link' for MeetingItem portal_types;
       8) Install every pending upgrades.
    '''
    migrator = Migrate_To_4105(context)
    migrator.run()
    migrator.finish()



# -*- coding: utf-8 -*-

from AccessControl import Unauthorized
from plone import api
from Products.Five.browser import BrowserView
from Products.PloneMeeting.browser.itemvotes import _get_linked_item_vote_numbers
from Products.PloneMeeting.config import NOT_ENCODED_VOTE_VALUE
from Products.PloneMeeting.config import PloneMeetingError
from Products.PloneMeeting.config import PMMessageFactory as _
from Products.PloneMeeting.utils import ItemPollTypeChangedEvent
from Products.PloneMeeting.utils import notifyModifiedAndReindex
from zope.component import queryUtility
from zope.event import notify
from zope.schema.interfaces import IVocabularyFactory


class ItemPollTypeView(BrowserView):
    '''Render the item pollType selection on the meetingitem_view.'''

    change_view_name = "change-item-polltype"
    js_onsuccess = "null"

    def __init__(self, context, request):
        super(BrowserView, self).__init__(context, request)
        self.context = context
        self.request = request
        self.portal = api.portal.get_tool('portal_url').getPortalObject()
        self.portal_url = self.portal.absolute_url()
        factory = queryUtility(IVocabularyFactory,
                               'Products.PloneMeeting.vocabularies.polltypesvocabulary')
        self.vocab = factory(self.context)

    def change_view_params(self, pollTypeTerm):
        """ """
        return {'new_value': pollTypeTerm.token}

    def selectablePollTypes(self):
        '''Returns a list of pollTypes the current user can set the item to.'''
        if not self.context.adapted().mayChangePollType():
            return []
        pollType = self.context.getPollType()
        return [term for term in self.vocab if term.value != pollType]


class ChangeItemPollTypeView(BrowserView):
    '''This manage the item pollType changes.'''

    def __init__(self, context, request):
        super(BrowserView, self).__init__(context, request)
        self.context = context
        self.request = request

    def __call__(self, new_value):
        '''Change pollType value.'''
        self._changePollType(new_value)

    def _do_validate_new_poll_type(self, old_pollType, new_value):
        """Specific validation."""
        # if user tries to switch from a public pollType to a secret
        # and vice-versa, it can not be done if some votes are encoded
        is_switching_vote_mode = (old_pollType.startswith('secret') and
                                  not new_value.startswith('secret')) or \
                                 (not old_pollType.startswith('secret') and
                                  new_value.startswith('secret'))
        if (new_value == 'no_vote' or is_switching_vote_mode) and \
           self.context.get_item_votes(include_unexisting=False):
            api.portal.show_message(
                _('can_not_switch_polltype_votes_encoded'),
                request=self.request,
                type='warning')
            return True

    def validate_new_poll_type(self, old_pollType, new_value):
        '''Make sure the new selected value can be selected.'''
        # make sure new_value exists
        factory = queryUtility(IVocabularyFactory,
                               'Products.PloneMeeting.vocabularies.polltypesvocabulary')
        if new_value not in factory(self.context):
            raise KeyError("New value '{0}' does not correspond to a value of MeetingItem.pollType".
                           format(new_value))

        if not self.context.adapted().mayChangePollType():
            raise Unauthorized

        # if common validation pass, call specific validation
        return self._do_validate_new_poll_type(old_pollType, new_value)

    def _changePollType(self, new_value):
        '''Helper method that changes pollType value and check that :
           - new_value is among selectable pollType values;
           - user actually mayChangePollType;
           - adapt Meeting.item_votes values.'''
        old_pollType = self.context.getPollType()
        if self.validate_new_poll_type(old_pollType, new_value):
            return

        # save old_pollType so we can pass it the the ItemPollTypeChangedEvent
        # set the new pollType and notify events
        self.context.setPollType(new_value)
        self.context._update_after_edit(idxs=['pollType'])
        try:
            notify(ItemPollTypeChangedEvent(self.context, old_pollType))
        except PloneMeetingError, msg:
            # back to original state
            self.context.setPollType(old_pollType)
            self.context._update_after_edit(idxs=['pollType'])
            api.portal.show_message(msg, type='warning')

        # an item's pollType has been changed, notify meeting
        if self.context.hasMeeting():
            meeting = self.context.getMeeting()
            notifyModifiedAndReindex(meeting)


class ItemVotePollTypeView(ItemPollTypeView):
    '''Render the item pollType selection on the item votes view.'''

    change_view_name = "change-item-vote-polltype"
    js_onsuccess = "onsuccessManageAttendees"

    def __call__(self, vote_number):
        ''' '''
        self.vote_number = vote_number
        return super(ItemVotePollTypeView, self).__call__()

    def change_view_params(self, pollTypeTerm):
        """ """
        params = super(ItemVotePollTypeView, self).change_view_params(pollTypeTerm)
        params['vote_number:int'] = self.vote_number
        return params

    def selectablePollTypes(self):
        '''Returns a list of pollTypes the current user can set the item to.'''
        if not self.context._mayChangeAttendees():
            return []
        pollType = self.context.get_item_votes(self.vote_number).get(
            'poll_type', self.context.getPollType())
        return [term for term in self.vocab
                if term.value not in [pollType, 'no_vote']]


class ChangeItemVotePollTypeView(ChangeItemPollTypeView):
    '''This manage the item vote pollType changes.'''

    def __call__(self, vote_number, new_value):
        '''Change pollType value.'''
        self.vote_number = vote_number
        return super(ChangeItemVotePollTypeView, self).__call__(new_value)

    def _do_validate_new_poll_type(self, old_pollType, new_value):
        """Specific validation."""
        # only if no vote encoded
        meeting = self.context.getMeeting()
        if self.context.get_vote_count(
            meeting, NOT_ENCODED_VOTE_VALUE, vote_number=self.vote_number) != \
           len(self.context.get_item_voters()):
            api.portal.show_message(
                _('can_not_switch_vote_polltype_votes_encoded'),
                request=self.request,
                type='warning')
            return True
        # can not be linked to other vote
        elif self.vote_number in _get_linked_item_vote_numbers(
                self.context, meeting, self.vote_number):
            api.portal.show_message(
                _('can_not_switch_vote_polltype_linked_to_previous'),
                request=self.request,
                type='warning')
            return True
        else:
            return False

    def _changePollType(self, new_value):
        ''' '''
        old_pollType = self.context.get_item_votes(self.vote_number).get(
            'poll_type', self.context.getPollType())
        if self.validate_new_poll_type(old_pollType, new_value):
            return
        # set new vote pollType
        meeting = self.context.getMeeting()
        if new_value.startswith('secret'):
            data = self.context._build_unexisting_vote(
                is_secret=True, vote_number=0, poll_type=new_value)[0]
            meeting.set_item_secret_vote(
                self.context, data, vote_number=self.vote_number)
        else:
            data = self.context._build_unexisting_vote(
                is_secret=False,
                vote_number=0,
                poll_type=new_value,
                voter_uids=self.context.get_item_voters())[0]
            meeting.set_item_public_vote(
                self.context, data, vote_number=self.vote_number)



# -*- coding: utf-8 -*-

from plone.app.workflow.remap import remap_workflow
from Products.PloneMeeting.migrations import logger
from Products.PloneMeeting.migrations import Migrator
from Products.ZCatalog.ProgressHandler import ZLogHandler


class Migrate_To_4106(Migrator):

    def _updateOrgsDashboardCollectionColumns(self):
        """Replace the 'pretty_link' column with the 'org_pretty_link_with_additional_infos'."""
        logger.info("Updating dashboard organization collections to replace the \"pretty_link\" "
                    "column by the\"org_pretty_link_with_additional_infos\" column...")
        hps_searches_folder = self.portal.contacts.get('hps-searches')
        persons_searches_folder = self.portal.contacts.get('persons-searches')
        # hide the 'review_state' column for orgs related collections
        for org_coll in hps_searches_folder.objectValues() + persons_searches_folder.objectValues():
            if 'org_pretty_link_with_additional_infos' not in org_coll.customViewFields:
                # make sure 'pretty_link' column no more present
                customViewFields = [col_name for col_name in org_coll.customViewFields
                                    if col_name != u'pretty_link']
                customViewFields.insert(1, u'org_pretty_link_with_additional_infos')
                org_coll.customViewFields = customViewFields
        logger.info('Done.')

    def _umarkCreationFlagForEveryItems(self):
        """ """
        brains = self.catalog(meta_type='MeetingItem')
        pghandler = ZLogHandler(steps=100)
        pghandler.init('Cleaning ftw.labels wrong annotations...', len(brains))
        pghandler.info("Unmarking _at_creation_flag for every items...")
        i = 0
        for brain in brains:
            i += 1
            pghandler.report(i)
            item = brain.getObject()
            item.unmarkCreationFlag()
        pghandler.finish()
        logger.info('Done.')

    def _remapContactsWorkflows(self):
        """Use 'plonemeeting_activity_managers_workflow' instead 'collective_contact_core_workflow'
           for person and held_position portal_types."""
        logger.info("Changing workflow for person and held_position portal_types...")
        person_wf = self.wfTool.getWorkflowsFor('person')[0]
        if person_wf.getId() != 'plonemeeting_activity_managers_workflow':
            remap_workflow(
                context=self.portal,
                type_ids=['person', 'held_position'],
                chain=['plonemeeting_activity_managers_workflow'],
                state_map={'active': 'active',
                           'deactivated': 'inactive'})
        logger.info('Done.')

    def run(self, from_migration_to_41=False):
        logger.info('Migrating to PloneMeeting 4106...')
        self._updateOrgsDashboardCollectionColumns()
        self._umarkCreationFlagForEveryItems()
        self._remapContactsWorkflows()


def migrate(context):
    '''This migration function will:

       1) Update dashboards displaying persons and held_positions;
       2) Unmark creation flag for every items;
       3) Change workflows for person and held_position portal_types.
    '''
    migrator = Migrate_To_4106(context)
    migrator.run()
    migrator.finish()



# -*- coding: utf-8 -*-
#
# File: Meeting.py
#
# GNU General Public License (GPL)
#

from AccessControl import ClassSecurityInfo
from App.class_init import InitializeClass
from appy.gen import No
from collections import OrderedDict
from collective.behavior.talcondition.utils import _evaluateExpression
from collective.contact.plonegroup.config import get_registry_organizations
from collective.eeafaceted.dashboard.utils import enableFacetedDashboardFor
from copy import deepcopy
from DateTime import DateTime
from imio.helpers.cache import cleanRamCacheFor
from imio.helpers.cache import invalidate_cachekey_volatile_for
from imio.helpers.security import fplog
from imio.prettylink.interfaces import IPrettyLink
from OFS.ObjectManager import BeforeDeleteException
from persistent.list import PersistentList
from persistent.mapping import PersistentMapping
from plone import api
from plone.app.querystring.querybuilder import queryparser
from plone.app.uuid.utils import uuidToCatalogBrain
from plone.memoize import ram
from Products.Archetypes.atapi import BooleanField
from Products.Archetypes.atapi import DateTimeField
from Products.Archetypes.atapi import IntegerField
from Products.Archetypes.atapi import OrderedBaseFolder
from Products.Archetypes.atapi import OrderedBaseFolderSchema
from Products.Archetypes.atapi import registerType
from Products.Archetypes.atapi import RichWidget
from Products.Archetypes.atapi import Schema
from Products.Archetypes.atapi import StringField
from Products.Archetypes.atapi import TextAreaWidget
from Products.Archetypes.atapi import TextField
from Products.Archetypes.event import ObjectEditedEvent
from Products.CMFCore.permissions import ModifyPortalContent
from Products.CMFCore.permissions import ReviewPortalContent
from Products.CMFCore.utils import _checkPermission
from Products.CMFDynamicViewFTI.browserdefault import BrowserDefaultMixin
from Products.CMFPlone.utils import base_hasattr
from Products.PloneMeeting.browser.itemchangeorder import _compute_value_to_add
from Products.PloneMeeting.browser.itemchangeorder import _is_integer
from Products.PloneMeeting.browser.itemchangeorder import _to_integer
from Products.PloneMeeting.browser.itemchangeorder import _use_same_integer
from Products.PloneMeeting.browser.itemvotes import clean_voters_linked_to
from Products.PloneMeeting.config import NOT_ENCODED_VOTE_VALUE
from Products.PloneMeeting.config import NOT_VOTABLE_LINKED_TO_VALUE
from Products.PloneMeeting.config import PMMessageFactory as _
from Products.PloneMeeting.config import PROJECTNAME
from Products.PloneMeeting.config import READER_USECASES
from Products.PloneMeeting.interfaces import IMeetingWorkflowActions
from Products.PloneMeeting.interfaces import IMeetingWorkflowConditions
# from Products.PloneMeeting.utils import getStatesBefore
from Products.PloneMeeting.utils import _addManagedPermissions
from Products.PloneMeeting.utils import _base_extra_expr_ctx
from Products.PloneMeeting.utils import addDataChange
from Products.PloneMeeting.utils import display_as_html
from Products.PloneMeeting.utils import displaying_available_items
from Products.PloneMeeting.utils import fieldIsEmpty
from Products.PloneMeeting.utils import forceHTMLContentTypeForEmptyRichFields
from Products.PloneMeeting.utils import get_annexes
from Products.PloneMeeting.utils import get_next_meeting
from Products.PloneMeeting.utils import getCustomAdapter
from Products.PloneMeeting.utils import getDateFromDelta
from Products.PloneMeeting.utils import getFieldVersion
from Products.PloneMeeting.utils import getWorkflowAdapter
from Products.PloneMeeting.utils import hasHistory
from Products.PloneMeeting.utils import isPowerObserverForCfg
from Products.PloneMeeting.utils import ItemDuplicatedFromConfigEvent
from Products.PloneMeeting.utils import MeetingLocalRolesUpdatedEvent
from Products.PloneMeeting.utils import rememberPreviousData
from Products.PloneMeeting.utils import set_field_from_ajax
from Products.PloneMeeting.utils import toHTMLStrikedContent
from Products.PloneMeeting.utils import transformAllRichTextFields
from Products.PloneMeeting.utils import updateAnnexesAccess
from Products.PloneMeeting.utils import validate_item_assembly_value
from zope.component import getMultiAdapter
from zope.event import notify
from zope.i18n import translate
from zope.interface import implements

import copy
import interfaces
import itertools
import logging
import os


__author__ = """Gaetan DELANNAY <gaetan.delannay@geezteem.com>, Gauthier BASTIEN
<g.bastien@imio.be>, Stephan GEULETTE <s.geulette@imio.be>"""
__docformat__ = 'plaintext'

logger = logging.getLogger('PloneMeeting')

# PloneMeetingError-related constants -----------------------------------------
BEFOREDELETE_ERROR = 'A BeforeDeleteException was raised by "%s" while ' \
    'trying to delete a meeting with id "%s"'


# Adapters ---------------------------------------------------------------------
class MeetingWorkflowConditions(object):
    '''Adapts a meeting to interface IMeetingWorkflowConditions.'''
    implements(IMeetingWorkflowConditions)
    security = ClassSecurityInfo()

    def __init__(self, meeting):
        self.context = meeting
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)

    security.declarePublic('mayAcceptItems')

    def mayAcceptItems(self):
        if self.context.query_state() in self.cfg.getMeetingStatesAcceptingItemsForMeetingManagers():
            return True

    security.declarePublic('mayPublish')

    def mayPublish(self):
        if _checkPermission(ReviewPortalContent, self.context):
            return True

    security.declarePublic('mayPublishDecisions')

    def mayPublishDecisions(self):
        '''Used when 'hide_decisions_when_under_writing' wfAdaptation is active.'''
        res = False
        # The user just needs the "Review portal content" permission on the object
        if _checkPermission(ReviewPortalContent, self.context):
            res = True
        return res

    security.declarePublic('mayFreeze')

    def mayFreeze(self):
        res = False
        if _checkPermission(ReviewPortalContent, self.context):
            res = True
        return res

    security.declarePublic('mayDecide')

    def mayDecide(self):
        res = False
        if _checkPermission(ReviewPortalContent, self.context):
            res = True
        return res

    security.declarePublic('mayCorrect')

    def mayCorrect(self, destinationState=None):
        '''See doc in interfaces.py.'''
        if not _checkPermission(ReviewPortalContent, self.context):
            return

        meeting_state = self.context.query_state()
        if meeting_state == 'closed':
            if self.tool.isManager(realManagers=True) or \
               'meetingmanager_correct_closed_meeting' in self.cfg.getWorkflowAdaptations():
                return True
            else:
                return No(_('closed_meeting_not_correctable_by_config'))

        return True

    security.declarePublic('mayClose')

    def mayClose(self):
        if _checkPermission(ReviewPortalContent, self.context):
            return True

    security.declarePublic('mayChangeItemsOrder')

    def mayChangeItemsOrder(self):
        res = True
        if not _checkPermission(ModifyPortalContent, self.context) or \
           self.context.query_state() in Meeting.meetingClosedStates:
            res = False
        return res


InitializeClass(MeetingWorkflowConditions)


class MeetingWorkflowActions(object):
    '''Adapts a meeting to interface IMeetingWorkflowActions.'''
    implements(IMeetingWorkflowActions)
    security = ClassSecurityInfo()

    def __init__(self, meeting):
        self.context = meeting
        self.tool = api.portal.get_tool('portal_plonemeeting')
        self.cfg = self.tool.getMeetingConfig(self.context)

    security.declarePrivate('initSequenceNumber')

    def initSequenceNumber(self):
        '''When a meeting is published (or frozen, depending on workflow
           adaptations), we attribute him a sequence number.'''
        if self.context.getMeetingNumber() != -1:
            return  # Already done.
        if self.cfg.getYearlyInitMeetingNumber():
            # I must reinit the meeting number to 0 if it is the first
            # meeting of this year.
            prev = self.context.getPreviousMeeting()
            if prev and \
               (prev.getDate().year() != self.context.getDate().year()):
                self.context.setMeetingNumber(1)
                self.cfg.setLastMeetingNumber(1)
                return
        # If we are here, we must simply increment the meeting number.
        meetingNumber = self.cfg.getLastMeetingNumber() + 1
        self.context.setMeetingNumber(meetingNumber)
        self.cfg.setLastMeetingNumber(meetingNumber)

    security.declarePrivate('doPublish')

    def doPublish(self, stateChange):
        '''When publishing the meeting, initialize the sequence number.'''
        self.initSequenceNumber()

    security.declarePrivate('doFreeze')

    def doFreeze(self, stateChange):
        '''When freezing the meeting, we initialize sequence number.'''
        self.initSequenceNumber()

    security.declarePrivate('doDecide')

    def doDecide(self, stateChange):
        ''' '''
        pass

    security.declarePrivate('doClose')

    def doClose(self, stateChange):
        ''' '''
        # Set the firstItemNumber
        unrestrictedMethodsView = getMultiAdapter((self.context, self.context.REQUEST),
                                                  name='pm_unrestricted_methods')
        self.context.setFirstItemNumber(unrestrictedMethodsView.findFirstItemNumber())
        self.context.updateItemReferences()
        # remove annex previews of every items if relevant
        if self.cfg.getRemoveAnnexesPreviewsOnMeetingClosure():
            # add logging message to fingerpointing log
            for item in self.context.get_items(ordered=True):
                annexes = get_annexes(item)
                if annexes:
                    for annex in annexes:
                        self.tool._removeAnnexPreviewFor(item, annex)
                extras = 'item={0} number_of_annexes={1}'.format(repr(item), len(annexes))
                fplog('remove_annex_previews', extras=extras)
            msg = translate(
                u"Preview of annexes were deleted upon meeting closure.",
                domain='PloneMeeting',
                context=self.context.REQUEST)
            api.portal.show_message(msg, request=self.context.REQUEST)

    security.declarePrivate('doPublish_decisions')

    def doPublish_decisions(self, stateChange):
        '''When the wfAdaptation 'hide_decisions_when_under_writing' is activated.'''
        pass

    security.declarePrivate('doBackToCreated')

    def doBackToCreated(self, stateChange):
        ''' '''
        pass

    security.declarePrivate('doBackToDecided')

    def doBackToDecided(self, stateChange):
        # Oups when closing a meeting we have updated the firsItemNumber
        # we need to reverse our action
        self.context.setFirstItemNumber(-1)

    security.declarePrivate('doBackToPublished')

    def doBackToPublished(self, stateChange):
        wfTool = api.portal.get_tool('portal_workflow')
        for item in self.context.get_items():
            if item.query_state() == 'itemfrozen':
                wfTool.doActionFor(item, 'backToItemPublished')
                if item.isLate():
                    wfTool.doActionFor(item, 'backToPresented')
                    # This way we "hide" again all late items.

    security.declarePrivate('doBackToDecisionsPublished')

    def doBackToDecisionsPublished(self, stateChange):
        '''When the wfAdaptation 'hide_decisions_when_under_writing' is activated.'''
        pass

    security.declarePrivate('doBackToFrozen')

    def doBackToFrozen(self, stateChange):
        pass


InitializeClass(MeetingWorkflowActions)

schema = Schema((

    StringField(
        name='title',
        widget=StringField._properties['widget'](
            visible=False,
            label='Title',
            label_msgid='PloneMeeting_label_title',
            i18n_domain='PloneMeeting',
        ),
        accessor="Title",
        searchable=True,
    ),
    DateTimeField(
        name='date',
        widget=DateTimeField._properties['widget'](
            label='Date',
            label_msgid='PloneMeeting_label_date',
            i18n_domain='PloneMeeting',
        ),
        required=True,
    ),
    DateTimeField(
        name='startDate',
        widget=DateTimeField._properties['widget'](
            condition="python: here.attribute_is_used('startDate') and not here.isTemporary()",
            minute_step=1,
            label='Startdate',
            label_msgid='PloneMeeting_label_startDate',
            i18n_domain='PloneMeeting',
        ),
        optional=True,
    ),
    DateTimeField(
        name='midDate',
        widget=DateTimeField._properties['widget'](
            condition="python: here.attribute_is_used('midDate') and not here.isTemporary()",
            minute_step=1,
            label='Middate',
            label_msgid='PloneMeeting_label_midDate',
            i18n_domain='PloneMeeting',
        ),
        optional=True,
    ),
    DateTimeField(
        name='endDate',
        widget=DateTimeField._properties['widget'](
            condition="python: here.attribute_is_used('endDate') and not here.isTemporary()",
            minute_step=1,
            label='Enddate',
            label_msgid='PloneMeeting_label_endDate',
            i18n_domain='PloneMeeting',
        ),
        optional=True,
    ),
    DateTimeField(
        name='approvalDate',
        widget=DateTimeField._properties['widget'](
            condition="python: here.attribute_is_used('approvalDate')",
            minute_step=1,
            label='Approvaldate',
            label_msgid='PloneMeeting_label_approvalDate',
            i18n_domain='PloneMeeting',
        ),
        optional=True,
    ),
    DateTimeField(
        name='convocationDate',
        widget=DateTimeField._properties['widget'](
            condition="python: here.attribute_is_used('convocationDate')",
            minute_step=1,
            label='Convocationdate',
            label_msgid='PloneMeeting_label_convocationDate',
            i18n_domain='PloneMeeting',
        ),
        optional=True,
    ),
    TextField(
        name='assembly',
        allowable_content_types="text/plain",
        widget=TextAreaWidget(
            condition="python: 'assembly' in here.shownAssemblyFields()",
            label_msgid="meeting_assembly",
            description="MeetingAssembly",
            description_msgid="assembly_meeting_descr",
            label_method="getLabelAssembly",
            label='Assembly',
            i18n_domain='PloneMeeting',
        ),
        default_content_type="text/plain",
        default_method="getDefaultAssembly",
        default_output_type="text/html",
        optional=True,
    ),
    TextField(
        name='assemblyExcused',
        allowable_content_types="text/plain",
        optional=True,
        widget=TextAreaWidget(
            condition="python: 'assemblyExcused' in here.shownAssemblyFields()",
            description="MeetingAssemblyExcused",
            description_msgid="assembly_excused_meeting_descr",
            label='Assemblyexcused',
            label_msgid='meeting_assemblyExcused',
            i18n_domain='PloneMeeting',
        ),
        default_output_type="text/html",
        default_content_type="text/plain",
    ),
    TextField(
        name='assemblyAbsents',
        allowable_content_types="text/plain",
        optional=True,
        widget=TextAreaWidget(
            condition="python: 'assemblyAbsents' in here.shownAssemblyFields()",
            description="MeetingAssemblyAbsents",
            description_msgid="assembly_absents_meeting_descr",
            label='Assemblyabsents',
            label_msgid='meeting_assemblyAbsents',
            i18n_domain='PloneMeeting',
        ),
        default_output_type="text/html",
        default_content_type="text/plain",
    ),
    TextField(
        name='assemblyGuests',
        allowable_content_types="text/plain",
        optional=True,
        widget=TextAreaWidget(
            condition="python: 'assemblyGuests' in here.shownAssemblyFields()",
            label='Assemblyguests',
            label_msgid='meeting_assemblyGuests',
            i18n_domain='PloneMeeting',
        ),
        default_output_type="text/html",
        default_content_type="text/plain",
    ),
    TextField(
        name='assemblyProxies',
        allowable_content_types="text/plain",
        optional=True,
        widget=TextAreaWidget(
            condition="python: 'assemblyProxies' in here.shownAssemblyFields()",
            label='Assemblyproxies',
            label_msgid='meeting_assemblyProxies',
            i18n_domain='PloneMeeting',
        ),
        default_output_type="text/html",
        default_content_type="text/plain",
    ),
    TextField(
        name='assemblyStaves',
        allowable_content_types="text/plain",
        optional=True,
        widget=TextAreaWidget(
            condition="python: 'assemblyStaves' in here.shownAssemblyFields()",
            label='Assemblystaves',
            label_msgid='meeting_assemblyStaves',
            i18n_domain='PloneMeeting',
        ),
        default_output_type="text/html",
        default_method="getDefaultAssemblyStaves",
        default_content_type="text/plain",
    ),
    TextField(
        name='signatures',
        allowable_content_types=('text/plain',),
        optional=True,
        widget=TextAreaWidget(
            condition="here/showSignatures",
            label_msgid="meeting_signatures",
            label='Signatures',
            i18n_domain='PloneMeeting',
        ),
        default_content_type='text/plain',
        default_method="getDefaultSignatures",
    ),
    StringField(
        name='place',
        widget=StringField._properties['widget'](
            condition="python: here.attribute_is_used('place')",
            label='Place',
            label_msgid='PloneMeeting_label_place',
            i18n_domain='PloneMeeting',
        ),
        optional=True,
        searchable=True,
    ),
    BooleanField(
        name='extraordinarySession',
        default=False,
        widget=BooleanField._properties['widget'](
            condition="python: here.attribute_is_used('extraordinarySession')",
            label='Extraordinarysession',
            label_msgid='PloneMeeting_label_extraordinarySession',
            i18n_domain='PloneMeeting',
        ),
        optional=True,
    ),
    TextField(
        name='inAndOutMoves',
        allowable_content_types=('text/html',),
        widget=RichWidget(
            condition="python: here.showMeetingManagerReservedField('inAndOutMoves')",
            description="InAndOutMoves",
            description_msgid="field_reserved_to_meeting_managers_descr",
            label_msgid="PloneMeeting_inAndOutMoves",
            label='Inandoutmoves',
            i18n_domain='PloneMeeting',
        ),
        default_content_type="text/html",
        default_output_type="text/x-html-safe",
        optional=True,
        searchable=True,
    ),
    TextField(
        name='notes',
        allowable_content_types=('text/html',),
        widget=RichWidget(
            condition="python: here.showMeetingManagerReservedField('notes')",
            description="Notes",
            description_msgid="field_reserved_to_meeting_managers_descr",
            label_msgid="PloneMeeting_notes",
            label='Notes',
            i18n_domain='PloneMeeting',
        ),
        default_content_type="text/html",
        default_output_type="text/x-html-safe",
        searchable=True,
        optional=True,
    ),
    TextField(
        name='observations',
        allowable_content_types=('text/html',),
        widget=RichWidget(
            condition="python: here.attribute_is_used('observations')",
            description="Observations",
            description_msgid="field_vieawable_by_everyone_descr",
            label_msgid="PloneMeeting_meetingObservations",
            label='Observations',
            i18n_domain='PloneMeeting',
        ),
        default_content_type="text/html",
        default_output_type="text/x-html-safe",
        searchable=True,
        optional=True,
    ),
    DateTimeField(
        name='preMeetingDate',
        widget=DateTimeField._properties['widget'](
            condition="python: here.attribute_is_used('preMeetingDate')",
            label='Premeetingdate',
            label_msgid='PloneMeeting_label_preMeetingDate',
            i18n_domain='PloneMeeting',
        ),
        optional=True,
    ),
    StringField(
        name='preMeetingPlace',
        widget=StringField._properties['widget'](
            condition="python: here.attribute_is_used('preMeetingPlace')",
            label='Premeetingplace',
            label_msgid='PloneMeeting_label_preMeetingPlace',
            i18n_domain='PloneMeeting',
        ),
        searchable=True,
        optional=True,
    ),
    TextField(
        name='preObservations',
        allowable_content_types=('text/html',),
        widget=RichWidget(
            condition="python: here.attribute_is_used('preObservations')",
            description_msgid="field_vieawable_by_everyone_descr",
            label='Preobservations',
            label_msgid='PloneMeeting_label_preObservations',
            i18n_domain='PloneMeeting',
        ),
        default_content_type="text/html",
        default_output_type="text/x-html-safe",
        searchable=True,
        optional=True,
    ),
    TextField(
        name='committeeObservations',
        allowable_content_types=('text/html',),
        widget=RichWidget(
            condition="python: here.attribute_is_used('committeeObservations')",
            description_msgid="field_vieawable_by_everyone_descr",
            label='Committeeobservations',
            label_msgid='PloneMeeting_label_committeeObservations',
            i18n_domain='PloneMeeting',
        ),
        default_content_type="text/html",
        default_output_type="text/x-html-safe",
        searchable=True,
        optional=True,
    ),
    TextField(
        name='votesObservations',
        allowable_content_types=('text/html',),
        widget=RichWidget(
            condition="python: here.attribute_is_used('votesObservations') and "
                      "here.adapted().show_votesObservations()",
            description_msgid="field_vieawable_by_everyone_once_meeting_decided_descr",
            label='Votesobservations',
            label_msgid='PloneMeeting_label_votesObservations',
            i18n_domain='PloneMeeting',
        ),
        default_content_type="text/html",
        default_output_type="text/x-html-safe",
        searchable=True,
        optional=True,
    ),
    TextField(
        name='publicMeetingObservations',
        allowable_content_types=('text/html',),
        widget=RichWidget(
            condition="python: here.attribute_is_used('publicMeetingObservations')",
            description_msgid="field_vieawable_by_everyone_descr",
            label='Publicmeetingobservations',
            label_msgid='PloneMeeting_label_publicMeetingObservations',
            i18n_domain='PloneMeeting',
        ),
        default_content_type="text/html",
        default_output_type="text/x-html-safe",
        searchable=True,
        optional=True,
    ),
    TextField(
        name='secretMeetingObservations',
        allowable_content_types=('text/html',),
        widget=RichWidget(
            condition="python: here.showMeetingManagerReservedField('secretMeetingObservations')",
            description="Secretmeetingobservations",
            description_msgid="field_reserved_to_meeting_managers_descr",
            label='Secretmeetingobservations',
            label_msgid='PloneMeeting_label_secretMeetingObservations',
            i18n_domain='PloneMeeting',
        ),
        default_content_type="text/html",
        default_output_type="text/x-html-safe",
        searchable=True,
        optional=True,
    ),
    TextField(
        name='authorityNotice',
        allowable_content_types=('text/html',),
        widget=RichWidget(
            condition="python: here.showMeetingManagerReservedField('authorityNotice')",
            description="AuthorityNotice",
            description_msgid="field_reserved_to_meeting_managers_descr",
            label='Authoritynotice',
            label_msgid='PloneMeeting_label_authorityNotice',
            i18n_domain='PloneMeeting',
        ),
        default_content_type="text/html",
        default_output_type="text/x-html-safe",
        searchable=True,
        optional=True,
    ),
    IntegerField(
        name='meetingNumber',
        default=-1,
        widget=IntegerField._properties['widget'](
            label='Meetingnumber',
            label_msgid='PloneMeeting_label_meetingNumber',
            i18n_domain='PloneMeeting',
        ),
        write_permission="Manage portal",
    ),
    IntegerField(
        name='firstItemNumber',
        default=-1,
        widget=IntegerField._properties['widget'](
            label='Firstitemnumber',
            label_msgid='PloneMeeting_label_firstItemNumber',
            i18n_domain='PloneMeeting',
        ),
        required=True,
        validators=('isInt', ),
        write_permission="Manage portal",
    ),
    StringField(
        name='meetingConfigVersion',
        widget=StringField._properties['widget'](
            label='Meetingconfigversion',
            label_msgid='PloneMeeting_label_meetingConfigVersion',
            i18n_domain='PloneMeeting',
        ),
        write_permission="Manage portal",
    ),
    DateTimeField(
        name='deadlinePublish',
        widget=DateTimeField._properties['widget'](
            condition="python: here.attribute_is_used('deadlinePublish') and not here.isTemporary()",
            label='Deadlinepublish',
            label_msgid='PloneMeeting_label_deadlinePublish',
            i18n_domain='PloneMeeting',
        ),
        optional=True,
    ),
    DateTimeField(
        name='deadlineFreeze',
        widget=DateTimeField._properties['widget'](
            condition="python: here.attribute_is_used('deadlineFreeze') and not here.isTemporary()",
            label='Deadlinefreeze',
            label_msgid='PloneMeeting_label_deadlineFreeze',
            i18n_domain='PloneMeeting',
        ),
        optional=True,
    ),

),
)

Meeting_schema = OrderedBaseFolderSchema.copy() + \
    schema.copy()


class Meeting(OrderedBaseFolder, BrowserDefaultMixin):
    """ A meeting made of items """
    security = ClassSecurityInfo()
    implements(interfaces.IMeeting)

    meta_type = 'Meeting'
    _at_rename_after_creation = True

    schema = Meeting_schema

    meetingClosedStates = ['closed']

    # declarePublic so it is callable in item view template
    # when the meeting is not viewable by the current user
    security.declarePublic('getAssembly')

    security.declarePublic('getPrettyLink')

    def getPrettyLink(self,
                      prefixed=False,
                      short=True,
                      showContentIcon=False,
                      isViewable=True,
                      notViewableHelpMessage=None,
                      appendToUrl='',
                      link_pattern=None):
        """Return the IPrettyLink version of the title."""
        adapted = IPrettyLink(self)
        tool = api.portal.get_tool('portal_plonemeeting')
        adapted.contentValue = tool.formatMeetingDate(self,
                                                      withHour=True,
                                                      prefixed=prefixed,
                                                      short=short)
        adapted.isViewable = adapted.isViewable and isViewable
        if notViewableHelpMessage is not None:
            adapted.notViewableHelpMessage = notViewableHelpMessage
        adapted.showContentIcon = showContentIcon
        adapted.appendToUrl = appendToUrl
        if link_pattern:
            adapted.link_pattern = link_pattern
        return adapted.getLink()

    security.declarePublic('getRawQuery')

    def getRawQuery(self, force_linked_items_query=False, **kwargs):
        """Override default getRawQuery to manage our own."""
        # available items?
        if displaying_available_items(self) and not force_linked_items_query:
            res = self._availableItemsQuery()
        else:
            res = [{'i': 'linkedMeetingUID',
                    'o': 'plone.app.querystring.operation.selection.is',
                    'v': self.UID()}, ]
        return res

    def _availableItemsQuery(self):
        '''Check docstring in IMeeting.'''
        meeting = self.getSelf()
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(meeting)
        if meeting.query_state() not in cfg.getMeetingStatesAcceptingItemsForMeetingManagers():
            # make sure the query returns nothing, add a dummy parameter
            return [{'i': 'UID',
                     'o': 'plone.app.querystring.operation.selection.is',
                     'v': 'dummy_unexisting_uid'}]
        res = [{'i': 'portal_type',
                'o': 'plone.app.querystring.operation.selection.is',
                'v': cfg.getItemTypeName()},
               {'i': 'review_state',
                'o': 'plone.app.querystring.operation.selection.is',
                'v': 'validated'},
               ]

        # before late state, accept items having any preferred meeting
        late_state = self.adapted().getLateState()
        if meeting.query_state() in self.getStatesBefore(late_state):
            # get items for which the getPreferredMeetingDate is lower or
            # equal to the date of this meeting (self)
            # a no preferred meeting item getPreferredMeetingDate is 1950/01/01
            res.append({'i': 'getPreferredMeetingDate',
                        'o': 'plone.app.querystring.operation.date.lessThan',
                        'v': meeting.getDate()})
        else:
            # after late state, only query items for which preferred meeting is self
            res.append({'i': 'getPreferredMeetingDate',
                        'o': 'plone.app.querystring.operation.date.between',
                        'v': (DateTime('2000/01/01'), meeting.getDate())})
        return res

    security.declarePublic('getSort_on')

    def getSort_on(self, force_linked_items_query=False):
        """ """
        if displaying_available_items(self) and not force_linked_items_query:
            return 'getProposingGroup'
        else:
            return 'getItemNumber'

    security.declarePublic('selectedViewFields')

    def selectedViewFields(self):
        """ """
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self)
        # some columns are displayed in the 'Purpose' column
        if displaying_available_items(self):
            visibleCols = cfg.getAvailableItemsListVisibleColumns()
        else:
            visibleCols = cfg.getItemsListVisibleColumns()
        itemsListVisibleColumns = [col for col in visibleCols if not col.startswith('static_')]
        itemsListVisibleColumns.insert(0, u'pretty_link')
        if not displaying_available_items(self):
            itemsListVisibleColumns.insert(0, u'getItemNumber')
            itemsListVisibleColumns.insert(0, u'listType')
        itemsListVisibleColumns.append(u'select_row')
        # selectedViewFields must return a list of tuple
        return [(elt, elt) for elt in itemsListVisibleColumns]

    def _get_all_redefined_attendees(self, by_persons=False, only_keys=True):
        """Returns a list of dicts."""
        itemNonAttendees = self.getItemNonAttendees(by_persons=by_persons)
        itemAbsents = self.get_item_absents(by_persons=by_persons)
        itemExcused = self.get_item_excused(by_persons=by_persons)
        itemSignatories = self.get_item_signatories(by_signatories=by_persons)
        if only_keys:
            redefined_item_attendees = itemNonAttendees.keys() + \
                itemAbsents.keys() + itemExcused.keys() + itemSignatories.keys()
        else:
            redefined_item_attendees = itemNonAttendees, itemAbsents, \
                itemExcused, itemSignatories
        return redefined_item_attendees

    def post_validate(self, REQUEST=None, errors=None):
        '''Validate attendees in post_validate as there is no field in schema for it.
           - an attendee may not be unselected if something is redefined for it on an item.'''
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self)

        if cfg.isUsingContacts() and not self.isTemporary():
            # removed attendees
            # REQUEST.form['meeting_attendees'] is like
            # ['muser_attendeeuid1_attendee', 'muser_attendeeuid2_excused']
            stored_attendees = self.get_all_used_held_positions(the_objects=False)
            meeting_attendees = [attendee.split('_')[1] for attendee
                                 in REQUEST.form.get('meeting_attendees', [])]
            removed_meeting_attendees = set(stored_attendees).difference(meeting_attendees)
            # attendees redefined on items
            redefined_item_attendees = self._get_all_redefined_attendees(by_persons=True)
            conflict_attendees = removed_meeting_attendees.intersection(
                redefined_item_attendees)
            if conflict_attendees:
                attendee_uid = tuple(removed_meeting_attendees)[0]
                attendee_brain = uuidToCatalogBrain(attendee_uid)
                errors['meeting_attendees'] = translate(
                    'can_not_remove_attendee_redefined_on_items',
                    mapping={'attendee_title': attendee_brain.get_full_title},
                    domain='PloneMeeting',
                    context=REQUEST)
            else:
                # removed voters
                stored_voters = self.get_voters()
                meeting_voters = [voter.split('_')[1] for voter
                                  in REQUEST.form.get('meeting_voters', [])]
                removed_meeting_voters = set(stored_voters).difference(meeting_voters)
                # public, voters are known
                item_votes = self.get_item_votes()
                voter_uids = []
                highest_secret_votes = 0
                for votes in item_votes.values():
                    for vote in votes:
                        if 'voters' in vote:
                            # public
                            voter_uids += [k for k, v in vote['voters'].items()
                                           if v != NOT_ENCODED_VOTE_VALUE]
                        else:
                            secret_votes = sum([v for k, v in vote['votes'].items()])
                            if secret_votes > highest_secret_votes:
                                highest_secret_votes = secret_votes
                voter_uids = list(set(voter_uids))
                conflict_voters = removed_meeting_voters.intersection(
                    voter_uids)
                if conflict_voters:
                    voter_uid = tuple(removed_meeting_voters)[0]
                    voter_brain = uuidToCatalogBrain(voter_uid)
                    errors['meeting_attendees'] = translate(
                        'can_not_remove_public_voter_voted_on_items',
                        mapping={'attendee_title': voter_brain.get_full_title},
                        domain='PloneMeeting',
                        context=REQUEST)
                elif highest_secret_votes > len(meeting_voters):
                    errors['meeting_attendees'] = translate(
                        'can_not_remove_secret_voter_voted_on_items',
                        domain='PloneMeeting',
                        context=REQUEST)

        return errors

    security.declarePrivate('validate_date')

    def validate_date(self, value):
        '''There can't be several meetings with the same date and hour.'''
        catalog = api.portal.get_tool('portal_catalog')
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self)
        # DateTime('2020-05-28 11:00') will result in DateTime('2020/05/28 11:00:00 GMT+0')
        # DateTime('2020/05/28 11:00') will result in DateTime('2020/05/28 11:00:00 GMT+2')
        # so make sure value uses "/" instead "-"
        value = value.replace("-", "/")
        otherMeetings = catalog(portal_type=cfg.getMeetingTypeName(), getDate=DateTime(value))
        if otherMeetings:
            for m in otherMeetings:
                if m.getObject() != self:
                    return translate('meeting_with_same_date_exists',
                                     domain='PloneMeeting',
                                     context=self.REQUEST)

    security.declarePrivate('validate_place')

    def validate_place(self, value):
        '''If "other" is selected, field "place_other" must contain
           something.'''
        rq = self.REQUEST
        if (value == 'other') and not rq.get('place_other', None):
            return translate('place_other_required',
                             domain='PloneMeeting',
                             context=rq)

    security.declarePrivate('validate_assembly')

    def validate_assembly(self, value):
        '''Validate the assembly field.'''
        if not validate_item_assembly_value(value):
            return translate('Please check that opening "[[" have corresponding closing "]]".',
                             domain='PloneMeeting',
                             context=self.REQUEST)

    security.declarePublic('displayStrikedAssembly')

    def displayStrikedAssembly(self):
        """ """
        return toHTMLStrikedContent(self.getAssembly())

    security.declarePublic('displaySignatures')

    def displaySignatures(self):
        """Display signatures as HTML, make sure lines added at end
           of signatures are displayed on screen correctly."""
        return display_as_html(self.getSignatures(), self)

    security.declarePublic('get_all_used_held_positions')

    def get_all_used_held_positions(self, include_new=False, the_objects=True):
        '''This will return every currently stored held_positions.
           If include_new=True, extra held_positions newly selected in the
           configuration are added.
           If p_the_objects=True, we return held_position objects, UID otherwise.
           '''
        # used Persons are held_positions stored in orderedContacts
        contacts = hasattr(self.aq_base, 'orderedContacts') and list(self.orderedContacts) or []
        if include_new:
            # now getOrderedContacts from MeetingConfig and append new contacts at the end
            # this is the case while adding new contact and editing existing meeting
            tool = api.portal.get_tool('portal_plonemeeting')
            cfg = tool.getMeetingConfig(self)
            selectable_contacts = cfg.getOrderedContacts()
            new_selectable_contacts = [c for c in selectable_contacts if c not in contacts]
            contacts = contacts + new_selectable_contacts

        if the_objects:
            # query held_positions
            catalog = api.portal.get_tool('portal_catalog')
            brains = catalog(UID=contacts)

            # make sure we have correct order because query was not sorted
            # we need to sort found brains according to uids
            def getKey(item):
                return contacts.index(item.UID)
            brains = sorted(brains, key=getKey)
            contacts = [brain.getObject() for brain in brains]
        return tuple(contacts)

    security.declarePublic('getDefaultAttendees')

    def getDefaultAttendees(self):
        '''The default attendees are the active held_positions
           with 'present' in defaults.'''
        res = []
        if self.checkCreationFlag():
            used_held_positions = self.get_all_used_held_positions(include_new=True)
            res = [held_pos.UID() for held_pos in used_held_positions
                   if held_pos.defaults and 'present' in held_pos.defaults]
        return res

    security.declarePublic('getDefaultSignatories')

    def getDefaultSignatories(self):
        '''The default signatories are the active held_positions
           with a defined signature_number.'''
        res = []
        if self.checkCreationFlag():
            used_held_positions = self.get_all_used_held_positions(include_new=True)
            res = [held_pos for held_pos in used_held_positions
                   if held_pos.defaults and 'present' in held_pos.defaults and held_pos.signature_number]
        return {signer.UID(): signer.signature_number for signer in res}

    security.declarePublic('getDefaultVoters')

    def getDefaultVoters(self):
        '''The default voters are the active held_positions
           with 'voter' in defaults.'''
        res = []
        if self.checkCreationFlag():
            used_held_positions = self.get_all_used_held_positions(include_new=True)
            res = [held_pos.UID() for held_pos in used_held_positions
                   if held_pos.defaults and 'voter' in held_pos.defaults]
        return res

    def _getContacts(self, contact_type=None, uids=None, theObjects=False):
        """ """
        res = []
        orderedContacts = getattr(self, 'orderedContacts', OrderedDict())
        if contact_type:
            for uid, infos in orderedContacts.items():
                if infos[contact_type] and (not uids or uid in uids):
                    res.append(uid)
        else:
            res = uids

        if theObjects:
            catalog = api.portal.get_tool('portal_catalog')
            brains = catalog(UID=res)
            res = [brain.getObject() for brain in brains]

            # keep correct order that was lost by catalog query
            def getKey(item):
                return self.orderedContacts.keys().index(item.UID())
            res = sorted(res, key=getKey)
        return tuple(res)

    security.declarePublic('get_attendees')

    def get_attendees(self, theObjects=False):
        '''Returns the attendees in this meeting.'''
        return self._getContacts('attendee', theObjects=theObjects)

    security.declarePublic('getExcused')

    def getExcused(self, theObjects=False):
        '''Returns the excused in this meeting.'''
        return self._getContacts('excused', theObjects=theObjects)

    security.declarePublic('getAbsents')

    def getAbsents(self, theObjects=False):
        '''Returns the absents in this meeting.'''
        return self._getContacts('absent', theObjects=theObjects)

    security.declarePublic('get_voters')

    def get_voters(self, uids=None, theObjects=False):
        '''Returns the voters in this meeting.'''
        voters = self._getContacts('voter', uids=uids, theObjects=theObjects)
        return voters

    security.declarePublic('get_signatories')

    def get_signatories(self, theObjects=False, by_signature_number=False):
        '''See docstring in previous method.'''
        signers = self._getContacts('signer', theObjects=theObjects)
        # order is important in case we have several same signature_number, the first win
        if theObjects:
            res = OrderedDict(
                [(signer, self.orderedContacts[signer.UID()]['signature_number'])
                 for signer in signers])
        else:
            res = OrderedDict(
                [(signer_uid, self.orderedContacts[signer_uid]['signature_number'])
                 for signer_uid in signers])
        if by_signature_number:
            # reverse res so when several same signature_number, the first win
            res = OrderedDict(reversed(res.items()))
            # keys are values, values are keys
            res = {v: k for k, v in res.items()}
        return dict(res)

    security.declarePublic('getReplacements')

    def getReplacements(self, theObjects=False):
        '''See docstring in previous method.'''
        replaced_uids = self._getContacts('replacement', theObjects=theObjects)
        return {replaced_uid: self.orderedContacts[replaced_uid]['replacement']
                for replaced_uid in replaced_uids}

    def _get_item_not_present(self, attr, by_persons=False):
        '''Return item not present (itemAbsents, itemExcused) by default the attr dict has the item UID
           as key and list of not_present as values but if 'p_by_persons' is True, the informations
           are returned with not_present held position as key and list of items as value.'''
        if by_persons:
            # values are now keys, concatenate a list of lists and remove duplicates
            keys = tuple(set(list(itertools.chain.from_iterable(attr.values()))))
            data = {}
            for key in keys:
                data[key] = [k for k, v in attr.items() if key in v]
        else:
            data = copy.deepcopy(attr.data)
        return data

    security.declarePublic('get_item_absents')

    def get_item_absents(self, by_persons=False):
        ''' '''
        return self._get_item_not_present(self.itemAbsents, by_persons=by_persons)

    security.declarePublic('get_item_excused')

    def get_item_excused(self, by_persons=False):
        ''' '''
        return self._get_item_not_present(self.itemExcused, by_persons=by_persons)

    security.declarePublic('getItemNonAttendees')

    def getItemNonAttendees(self, by_persons=False):
        ''' '''
        return self._get_item_not_present(self.itemNonAttendees, by_persons=by_persons)

    security.declarePublic('get_item_signatories')

    def get_item_signatories(self, by_signatories=False, include_position_type=False):
        '''Return itemSignatories, by default the itemSignatories dict has the item UID as key and list
           of signatories as values but if 'p_by_signatories' is True, the informations are returned with
           signatory as key and list of items as value.
           p_include_position_type=True is only relevant when p_by_signatories=False.'''
        signatories = {}
        if by_signatories:
            for item_uid, signatories_infos in self.itemSignatories.items():
                for signature_number, signatory_infos in signatories_infos.items():
                    # do not keep 'position_type' from the stored itemSignatories
                    signatory_uid = signatory_infos['hp_uid']
                    if signatory_uid not in signatories:
                        signatories[signatory_uid] = []
                    signatories[signatory_uid].append(item_uid)
        else:
            for item_uid, signatory_infos in self.itemSignatories.data.items():
                if include_position_type:
                    signatories[item_uid] = signatory_infos.copy()
                else:
                    signatories[item_uid] = {k: v['hp_uid'] for k, v in signatory_infos.items()}

        return signatories

    def get_signature_infos_for(self,
                                item_uid,
                                signatory_uid,
                                render_position_type=False,
                                prefix_position_type=False):
        """Return the signature position_type to use as label and signature_number
           for given p_item_uid and p_signatory_uid."""
        # check if signatory_uid is redefined on the item
        data = self.get_item_signatories(by_signatories=False, include_position_type=True)
        data = {k: v for k, v in data[item_uid].items()
                if v['hp_uid'] == signatory_uid}
        catalog = api.portal.get_tool('portal_catalog')
        hp = catalog(UID=signatory_uid)[0].getObject()
        if data:
            signature_number, position_type = data.items()[0]
        else:
            # if not, then get it from meeting signatories
            signature_number = self.getSignatories()[signatory_uid]
            # position type is the one of the signatory (signatory_uid)
            position_type = hp.position_type
        res = {}
        res['signature_number'] = signature_number
        if render_position_type:
            if prefix_position_type:
                res['position_type'] = hp.get_prefix_for_gender_and_number(
                    include_value=True,
                    forced_position_type_value=position_type)
            else:
                res['position_type'] = hp.get_label(
                    forced_position_type_value=position_type)
        else:
            res['position_type'] = position_type
        return res

    security.declarePublic('get_item_votes')

    def get_item_votes(self):
        ''' '''
        votes = deepcopy(self.itemVotes.data)
        return votes

    security.declarePrivate('setItemPublicVote')

    def setItemPublicVote(self, item, data, vote_number=0):
        """ """
        data = deepcopy(data)
        item_uid = item.UID()
        # set new itemVotes value on meeting
        # first votes
        if item_uid not in self.itemVotes:
            self.itemVotes[item_uid] = PersistentList()
            # check if we are not adding a new vote on an item containing no votes at all
            if vote_number == 1:
                # add an empty vote 0
                data_item_vote_0 = item.get_item_votes(
                    vote_number=0,
                    include_vote_number=False,
                    include_unexisting=True)
                # make sure we use persistent for 'voters'
                data_item_vote_0['voters'] = PersistentMapping(data_item_vote_0['voters'])
                self.itemVotes[item_uid].append(PersistentMapping(data_item_vote_0))
        new_voters = data.get('voters')
        # new vote_number
        if vote_number + 1 > len(self.itemVotes[item_uid]):
            # complete data before storing, if some voters are missing it is
            # because of NOT_VOTABLE_LINKED_TO_VALUE, we add it
            item_voter_uids = item.get_item_voters()
            for item_voter_uid in item_voter_uids:
                if item_voter_uid not in data['voters']:
                    data['voters'][item_voter_uid] = NOT_VOTABLE_LINKED_TO_VALUE
            self.itemVotes[item_uid].append(PersistentMapping(data))
        else:
            # use update in case we only update a subset of votes
            # when some vote NOT_VOTABLE_LINKED_TO_VALUE or so
            # we have nested dicts, data is a dict, containing 'voters' dict
            self.itemVotes[item_uid][vote_number]['voters'].update(data['voters'])
            data.pop('voters')
            self.itemVotes[item_uid][vote_number].update(data)
        # manage linked_to_previous
        # if current vote is linked to other votes, we will set NOT_VOTABLE_LINKED_TO_VALUE
        # as value of vote of voters of other linked votes
        clean_voters_linked_to(item, self, vote_number, new_voters)

    security.declarePrivate('setItemSecretVote')

    def setItemSecretVote(self, item, data, vote_number):
        """ """
        data = deepcopy(data)
        item_uid = item.UID()
        # set new itemVotes value on meeting
        # first votes
        if item_uid not in self.itemVotes:
            self.itemVotes[item_uid] = PersistentList()
            # check if we are not adding a new vote on an item containing no votes at all
            if vote_number == 1:
                # add an empty vote 0
                data_item_vote_0 = item.get_item_votes(
                    vote_number=0,
                    include_vote_number=False,
                    include_unexisting=True)
                self.itemVotes[item_uid].append(PersistentMapping(data_item_vote_0))
        # new vote_number
        if vote_number + 1 > len(self.itemVotes[item_uid]):
            self.itemVotes[item_uid].append(PersistentMapping(data))
        else:
            self.itemVotes[item_uid][vote_number].update(data)

    security.declarePublic('displayUserReplacement')

    def displayUserReplacement(self,
                               held_position_uid,
                               include_held_position_label=True,
                               include_sub_organizations=True):
        '''Display the user remplacement from p_held_position_uid.'''
        catalog = api.portal.get_tool('portal_catalog')
        held_position = catalog(UID=held_position_uid)[0].getObject()
        if include_held_position_label:
            return held_position.get_short_title(
                include_sub_organizations=include_sub_organizations)
        else:
            person = held_position.get_person()
            return person.get_title()

    security.declarePrivate('setDate')

    def setDate(self, value, **kwargs):
        '''Overrides the field 'date' mutator so we reindex every linked
           items if date value changed.  Moreover we manage updateItemReferences
           if value changed.'''
        current_date = self.getField('date').get(self, **kwargs)
        if not value == current_date:
            # add a value in the REQUEST to specify that updateItemReferences is needed
            self.REQUEST.set('need_Meeting_updateItemReferences', True)
            # store new date before updating items so items get
            # right date when calling meeting.getDate
            self.getField('date').set(self, value, **kwargs)
            catalog = api.portal.get_tool('portal_catalog')
            tool = api.portal.get_tool('portal_plonemeeting')
            cfg = tool.getMeetingConfig(self)
            # items linked to the meeting
            brains = catalog(portal_type=cfg.getItemTypeName(),
                             linkedMeetingUID=self.UID())
            # items having the meeting as the preferredMeeting
            brains = brains + catalog(portal_type=cfg.getItemTypeName(),
                                      getPreferredMeeting=self.UID())
            for brain in brains:
                item = brain.getObject()
                item.reindexObject(idxs=['linkedMeetingDate', 'getPreferredMeetingDate'])
            # clean cache for "Products.PloneMeeting.vocabularies.meetingdatesvocabulary"
            invalidate_cachekey_volatile_for(
                "Products.PloneMeeting.vocabularies.meetingdatesvocabulary", get_again=True)

    security.declarePrivate('setFirstItemNumber')

    def setFirstItemNumber(self, value, **kwargs):
        '''Overrides the field 'firstItemNumber' mutator to be able to
           updateItemReferences if value changed.'''
        current_first_item_number = self.getField('firstItemNumber').get(self, **kwargs)
        if not value == current_first_item_number:
            # add a value in the REQUEST to specify that updateItemReferences is needed
            self.REQUEST.set('need_Meeting_updateItemReferences', True)
        self.getField('firstItemNumber').set(self, value, **kwargs)

    security.declarePrivate('setMeetingNumber')

    def setMeetingNumber(self, value, **kwargs):
        '''Overrides the field 'meetingNumber' mutator to be able to
           updateItemReferences if value changed.'''
        current_meetingNumber = self.getField('meetingNumber').get(self, **kwargs)
        if not value == current_meetingNumber:
            # add a value in the REQUEST to specify that updateItemReferences is needed
            self.REQUEST.set('need_Meeting_updateItemReferences', True)
        self.getField('meetingNumber').set(self, value, **kwargs)

    security.declarePublic('showMeetingManagerReservedField')

    def showMeetingManagerReservedField(self, name):
        '''When must field named p_name be shown?'''
        tool = api.portal.get_tool('portal_plonemeeting')
        res = not self.isTemporary() and \
            tool.isManager(self) and \
            self.attribute_is_used(name)
        return res

    security.declarePrivate('validate_preMeetingDate')

    def validate_preMeetingDate(self, value):
        '''Checks that the preMeetingDate comes before the meeting date.'''
        if not value or not self.attribute_is_used('preMeetingDate'):
            return
        # Get the meeting date from the request
        try:
            meetingDate = DateTime(self.REQUEST['date'])
        except DateTime.DateError:
            meetingDate = None
        except DateTime.SyntaxError:
            meetingDate = None
        # Compare meeting and pre-meeting dates
        if meetingDate and (DateTime(value) >= meetingDate):
            label = 'pre_date_after_meeting_date'
            return translate(label, domain='PloneMeeting', context=self.REQUEST)

    def get_items(self,
                 uids=[],
                 listTypes=[],
                 ordered=False,
                 theObjects=True,
                 additional_catalog_query={},
                 unrestricted=False,
                 force_linked_items_query=True):
        '''Overrides the Meeting.items accessor.
           Items can be filtered depending on :
           - list of given p_uids;
           - given p_listTypes;
           - returned ordered (by getItemNumber) if p_ordered is True;
           - if p_theObjects is True, MeetingItem objects are returned, else, brains are returned;
           - if p_unrestricted is True it will return every items, not checking permission;
           - if p_force_linked_items_query is True, it will call self.getRawQuery with
             same parameter and force use of query showing linked items, not displaying
             available items.
        '''
        # execute the query using the portal_catalog
        catalog = api.portal.get_tool('portal_catalog')
        catalog_query = self.getRawQuery(force_linked_items_query=force_linked_items_query)
        if listTypes:
            catalog_query.append({'i': 'listType',
                                  'o': 'plone.app.querystring.operation.selection.is',
                                  'v': listTypes},)
        if uids:
            catalog_query.append({'i': 'UID',
                                  'o': 'plone.app.querystring.operation.selection.is',
                                  'v': uids},)
        if ordered:
            query = queryparser.parseFormquery(
                self,
                catalog_query,
                sort_on=self.getSort_on(force_linked_items_query=force_linked_items_query))
        else:
            query = queryparser.parseFormquery(self, catalog_query)

        # append additional_catalog_query
        query.update(additional_catalog_query)
        if unrestricted:
            res = catalog.unrestrictedSearchResults(**query)
        else:
            res = catalog(**query)

        if theObjects:
            res = [brain._unrestrictedGetObject() for brain in res]
        return res

    def getRawItems(self):
        """Simply get linked items."""
        catalog = api.portal.get_tool('portal_catalog')
        catalog_query = self.getRawQuery(force_linked_items_query=True)
        query = queryparser.parseFormquery(self, catalog_query)
        res = [brain.UID for brain in catalog.unrestrictedSearchResults(**query)]
        return res

    security.declarePublic('getItemsInOrder')

    def getItemsInOrder(self, late=False, uids=[]):
        """Deprecated, use Meeting.getItems instead."""
        logger.warn('Meeting.getItemsInOrder is deprecated, use Meeting.get_items(ordered=True) instead.')
        listTypes = late and ['late'] or ['normal']
        if '' in uids:
            uids.remove('')
        return self.get_items(uids=uids, listTypes=listTypes, ordered=True)

    security.declarePublic('getItemByNumber')

    def getItemByNumber(self, number):
        '''Gets the item thas has number p_number.'''
        catalog = api.portal.get_tool('portal_catalog')
        brains = catalog(linkedMeetingUID=self.UID(), getItemNumber=number)
        if not brains:
            return None
        return brains[0].getObject()

    def getLateState(self):
        '''See doc in interfaces.py.'''
        return 'frozen'

    def getStatesBefore_cachekey(method, self, review_state):
        '''cachekey method for self.getStatesBefore.'''
        # do only re-compute if cfg changed or params changed
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self)
        return (cfg.getId(), cfg._p_mtime, review_state)

    @ram.cache(getStatesBefore_cachekey)
    def getStatesBefore(self, review_state):
        """
          Returns states before the p_review_state state.
        """
        return getStatesBefore(self, review_state)

    def _check_insert_order_cache(self, cfg):
        '''See doc in interfaces.py.'''
        meeting = self.getSelf()
        invalidate = False
        invalidate = not base_hasattr(meeting, '_insert_order_cache') or \
            meeting._insert_order_cache['categories_modified'] != cfg.categories.modified() or \
            meeting._insert_order_cache['plonegroup_orgs'] != get_registry_organizations()

        # check cfg attrs
        if not invalidate:
            for key in meeting._insert_order_cache.keys():
                if key.startswith('cfg_'):
                    if meeting._insert_order_cache[key] != getattr(cfg, key[4:]):
                        invalidate = True
                        break
        if invalidate:
            meeting.adapted()._init_insert_order_cache(cfg)
        return invalidate

    def _insert_order_cache_cfg_attrs(self, cfg):
        '''See doc in interfaces.py.'''
        return ['insertingMethodsOnAddItem',
                'listTypes',
                'selectablePrivacies',
                'usedPollTypes',
                'orderedAssociatedOrganizations',
                'orderedGroupsInCharge']

    def _init_insert_order_cache(self, cfg):
        '''See doc in interfaces.py.'''
        meeting = self.getSelf()
        meeting._insert_order_cache = PersistentMapping()
        for cfg_attr in meeting.adapted()._insert_order_cache_cfg_attrs(cfg):
            key = 'cfg_{0}'.format(cfg_attr)
            value = deepcopy(getattr(cfg, cfg_attr))
            meeting._insert_order_cache[key] = value
        meeting._insert_order_cache['categories_modified'] = cfg.categories.modified()
        meeting._insert_order_cache['plonegroup_orgs'] = get_registry_organizations()
        meeting._insert_order_cache['items'] = PersistentMapping()

    def _invalidate_insert_order_cache_for(self, item):
        '''Invalidate cache for given p_item.'''
        item_uid = item.UID()
        if base_hasattr(self, '_insert_order_cache') and \
           self._insert_order_cache['items'].get(item_uid, None) is not None:
            del self._insert_order_cache['items'][item_uid]

    def getItemInsertOrder(self, item, cfg, check_cache=True):
        '''Get p_item insertOrder taking cache into account.'''
        # check if cache still valid, will be invalidated if not
        if check_cache:
            self.adapted()._check_insert_order_cache(cfg)
        insert_order = self._insert_order_cache['items'].get(item.UID(), None)
        if insert_order is None or not isinstance(insert_order, list):
            insert_order = item.adapted()._getInsertOrder(cfg)
            self._insert_order_cache['items'][item.UID()] = insert_order
        return insert_order

    security.declareProtected(ModifyPortalContent, 'insertItem')

    def insertItem(self, item, forceNormal=False):
        '''Inserts p_item into my list of "normal" items or my list of "late"
           items. If p_forceNormal is True, and the item should be inserted as
           a late item, it is nevertheless inserted as a normal item.'''
        # First, determine if we must insert the item into the "normal"
        # list of items or to the list of "late" items. Note that I get
        # the list of items *in order* in the case I need to insert the item
        # at another place than at the end.
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self)
        isLate = not forceNormal and item.wfConditions().isLateFor(self)
        if isLate:
            item.setListType(item.adapted().getListTypeLateValue(self))
            toDiscussValue = cfg.getToDiscussLateDefault()
        else:
            item.setListType(item.adapted().getListTypeNormalValue(self))
            toDiscussValue = cfg.getToDiscussDefault()
        items = self.get_items(ordered=True)
        # Set the correct value for the 'toDiscuss' field if required
        if cfg.getToDiscussSetOnItemInsert():
            item.setToDiscuss(toDiscussValue)
        # At what place must we insert the item in the list ?
        insertMethods = cfg.getInsertingMethodsOnAddItem()
        # wipe out insert methods as stored value is a DataGridField
        # and we only need a tuple of insert methods
        insertAtTheEnd = False
        if insertMethods[0]['insertingMethod'] != 'at_the_end':
            # We must insert it according to category or proposing group order
            # (at the end of the items belonging to the same category or
            # proposing group). We will insert the p_item just before the first
            # item whose category/group immediately follows p_item's category/
            # group (or at the end if inexistent). Note that the MeetingManager,
            # in subsequent manipulations, may completely change items order.
            self.adapted()._check_insert_order_cache(cfg)
            itemOrder = self.getItemInsertOrder(item, cfg, check_cache=False)
            higherItemFound = False
            insertIndex = 0  # That's where I will insert the item
            insertIndexIsSubnumber = False
            for anItem in items:
                if higherItemFound:
                    itemNumber = anItem.getItemNumber()
                    # Ok I already know where to insert the item. I just
                    # continue to visit the next items in order to increment their number.
                    # we inserted an integer numer, we need to add '1' to every next items
                    if not insertIndexIsSubnumber:
                        anItem.setItemNumber(itemNumber + 100)
                    elif (insertIndexIsSubnumber and _use_same_integer(itemNumber, insertIndex) and
                          itemNumber > insertIndex):
                        # we inserted a subnumber, we need to update subnumber of same integer
                        anItem.setItemNumber(itemNumber + 1)
                elif self.getItemInsertOrder(anItem, cfg, check_cache=False) > itemOrder:
                    higherItemFound = True
                    itemNumber = anItem.getItemNumber()
                    insertIndex = itemNumber
                    # we will only update next items of same subnumber?
                    insertIndexIsSubnumber = not _is_integer(itemNumber)
                    anItem.setItemNumber(itemNumber + _compute_value_to_add(itemNumber))

            if higherItemFound:
                item.setItemNumber(insertIndex)
            else:
                insertAtTheEnd = True

        if insertMethods[0]['insertingMethod'] == 'at_the_end' or insertAtTheEnd:
            # insert it as next integer number
            if items:
                item.setItemNumber(_to_integer(items[-1].getItemNumber()) + 100)
            else:
                # first added item
                item.setItemNumber(100)

        # Add the item at the end of the items list
        item._update_meeting_link(self)
        self._finalize_item_insert(items_to_update=[item])

    def _finalize_item_insert(self, items_to_update=[]):
        """ """
        # invalidate RAMCache for MeetingItem.getMeeting
        cleanRamCacheFor('Products.PloneMeeting.MeetingItem.getMeeting')
        # reindex getItemNumber when item is in the meeting or getItemNumber returns None
        # and reindex linkedMeeting indexes that is used by updateItemReferences using getItems
        lowest_item_number = 0
        for item in items_to_update:
            itemNumber = item.getRawItemNumber()
            if not lowest_item_number or itemNumber < lowest_item_number:
                lowest_item_number = itemNumber
            item.reindexObject(idxs=['getItemNumber',
                                     'listType',
                                     'linkedMeetingUID',
                                     'linkedMeetingDate'])
        # meeting is considered modified, do this before updateItemReferences
        self.notifyModified()

        # update itemReference after 'getItemNumber' has been reindexed of item and
        # items with a higher itemNumber
        self.updateItemReferences(startNumber=lowest_item_number)

    security.declareProtected(ModifyPortalContent, 'removeItem')

    def removeItem(self, item):
        '''Removes p_item from me.'''
        # Remember the item number now; once the item will not be in the meeting
        # anymore, it will loose its number.
        itemNumber = item.getItemNumber()
        items = self.get_items()
        try:
            item._update_meeting_link(None)
            items.remove(item)
            # set listType back to 'normal' if it was late
            # if it is another value (custom), we do not change it
            if item.getListType() == 'late':
                item.setListType('normal')
        except ValueError:
            # in case this is called by onItemRemoved, the item
            # does not exist anymore and is no more in the items list
            # so we pass
            pass

        # remove item UID from self.itemAbsents/self.itemExcused and self.itemSignatories
        item_uid = item.UID()
        if item_uid in self.itemAbsents:
            del self.itemAbsents[item_uid]
        if item_uid in self.itemExcused:
            del self.itemExcused[item_uid]
        if item_uid in self.itemNonAttendees:
            del self.itemNonAttendees[item_uid]
        if item_uid in self.itemSignatories:
            del self.itemSignatories[item_uid]
        if item_uid in self.itemVotes:
            del self.itemVotes[item_uid]

        # remove item UID from _insert_order_cache
        self._invalidate_insert_order_cache_for(item)

        # make sure item assembly/signatures related fields are emptied
        for field in item.Schema().filterFields(isMetadata=False):
            if field.getName().startswith('itemAssembly') or field.getName() == 'itemSignatures':
                field.set(item, '')

        # Update item numbers
        # in case itemNumber was a subnumber (or a master having subnumber),
        # we will just update subnumbers of the same integer
        itemNumberIsSubnumber = not _is_integer(itemNumber) or bool(self.getItemByNumber(itemNumber + 1))
        for anItem in items:
            anItemNumber = anItem.getItemNumber()
            if anItemNumber > itemNumber:
                if not itemNumberIsSubnumber:
                    anItem.setItemNumber(anItem.getItemNumber() - 100)
                elif itemNumberIsSubnumber and _use_same_integer(itemNumber, anItemNumber):
                    anItem.setItemNumber(anItem.getItemNumber() - _compute_value_to_add(anItemNumber))
        # invalidate RAMCache for MeetingItem.getMeeting
        cleanRamCacheFor('Products.PloneMeeting.MeetingItem.getMeeting')

        # reindex relevant indexes now that item is removed
        item.reindexObject(idxs=['listType', 'linkedMeetingUID', 'linkedMeetingDate'])

        # meeting is considered modified, do this before updateItemReferences
        self.notifyModified()

        # update itemReference of item that is no more linked to self and so that will not
        # be updated by Meeting.updateItemReferences and then update items that used
        # a higher item number
        item.updateItemReference()
        self.updateItemReferences(startNumber=itemNumber)

    def updateItemReferences(self, startNumber=0, check_needed=False):
        """Update itemReference of every contained items, if p_startNumber is given,
           we update items starting from p_startNumber itemNumber.
           By default, if p_startNumber=0, every linked items will be updated.
           If p_check_needed is True, we check if value 'need_Meeting_updateItemReferences' in REQUEST is True."""
        # call to updateItemReferences may be deferred for optimization
        if self.REQUEST.get('defer_Meeting_updateItemReferences', False):
            return
        if check_needed and not self.REQUEST.get('need_Meeting_updateItemReferences', False):
            return
        # force disable 'need_Meeting_updateItemReferences' from REQUEST
        self.REQUEST.set('need_Meeting_updateItemReferences', False)

        # we query items from startNumber to last item of the meeting
        # moreover we getItems unrestricted to be sure we have every elements
        brains = self.get_items(
            ordered=True,
            theObjects=False,
            unrestricted=True,
            additional_catalog_query={
                'getItemNumber': {'query': startNumber,
                                  'range': 'min'}, })
        for brain in brains:
            item = brain._unrestrictedGetObject()
            item.updateItemReference()

    security.declarePrivate('getDefaultAssembly')

    def getDefaultAssembly(self):
        if self.attribute_is_used('assembly'):
            tool = api.portal.get_tool('portal_plonemeeting')
            return tool.getMeetingConfig(self).getAssembly()
        return ''

    security.declarePrivate('getDefaultAssemblyStaves')

    def getDefaultAssemblyStaves(self):
        if self.attribute_is_used('assemblyStaves'):
            tool = api.portal.get_tool('portal_plonemeeting')
            return tool.getMeetingConfig(self).getAssemblyStaves()
        return ''

    security.declarePrivate('getDefaultSignatures')

    def getDefaultSignatures(self):
        if self.attribute_is_used('signatures'):
            tool = api.portal.get_tool('portal_plonemeeting')
            cfg = tool.getMeetingConfig(self)
            return cfg.getSignatures()
        return ''

    security.declarePrivate('updateTitle')

    def updateTitle(self):
        '''The meeting title is generated by this method, based on the meeting date.'''
        tool = api.portal.get_tool('portal_plonemeeting')
        title = self.title
        formatted_date = tool.formatMeetingDate(self)
        if title != formatted_date:
            self.setTitle(formatted_date)
            return True

    security.declarePrivate('updatePlace')

    def updatePlace(self):
        '''Updates the place if it comes from special request field
           "place_other".'''
        rq = self.REQUEST
        if ('place' not in rq) or (rq.get('place', '') == 'other'):
            stored_place = self.getPlace()
            new_place = rq.get('place_other', '')
            if stored_place != new_place:
                self.setPlace(new_place)
                return True

    security.declarePrivate('computeDates')

    def computeDates(self):
        '''Computes, for this meeting, the dates which are derived from the
           meeting date when relevant.'''
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self)
        usedAttrs = cfg.getUsedMeetingAttributes()
        meetingDate = self.getDate()
        # Initialize the effective start date with the meeting date
        if 'startDate' in usedAttrs:
            self.setStartDate(meetingDate)
        # Set, by default, mid date to start date + 1 hour.
        if 'midDate' in usedAttrs:
            self.setMidDate(meetingDate + 1 / 24.0)
        # Set, by default, end date to start date + 2 hours.
        if 'endDate' in usedAttrs:
            self.setEndDate(meetingDate + 2 / 24.0)
        # Compute the publish deadline
        if 'deadlinePublish' in usedAttrs and not self.getDeadlinePublish():
            delta = cfg.getPublishDeadlineDefault()
            if not delta.strip() in ('', '0',):
                self.setDeadlinePublish(getDateFromDelta(meetingDate, '-' + delta))
        if 'deadlineFreeze' in usedAttrs and not self.getDeadlineFreeze():
            # Compute the freeze deadline
            delta = cfg.getFreezeDeadlineDefault()
            if not delta.strip() in ('', '0',):
                self.setDeadlineFreeze(getDateFromDelta(meetingDate, '-' + delta))
        if 'preMeetingDate' in usedAttrs and not self.getPreMeetingDate():
            # Compute the date for the pre-meeting
            delta = cfg.getPreMeetingDateDefault()
            if not delta.strip() in ('', '0',):
                self.setPreMeetingDate(getDateFromDelta(meetingDate, '-' + delta))

    security.declarePublic('getUserReplacements')

    def getUserReplacements(self):
        '''Gets the dict storing user replacements.'''
        res = {}
        orderedContacts = getattr(self, 'orderedContacts', OrderedDict())
        for uid, infos in orderedContacts.items():
            if infos['replacement']:
                res[uid] = infos['replacement']
        return res

    security.declarePublic('filterPossibleUserReplacement')

    def filterPossibleUserReplacement(self, allUsers):
        '''Adaptable method to filter possible user replacement.'''
        return allUsers

    def _doUpdateContacts(self,
                          attendees=OrderedDict(),
                          signatories={},
                          replacements={},
                          voters=[]):
        ''' '''
        # attendees must be an OrderedDict to keep order
        if not isinstance(attendees, OrderedDict):
            raise ValueError(
                'Parameter attendees passed to Meeting._doUpdateContacts '
                'must be an OrderedDict !!!')
        # save the ordered contacts so we rely on this, especially when
        # users are disabled in the configuration
        self.orderedContacts.clear()

        for attendee_uid, attendee_type in attendees.items():
            if attendee_uid not in self.orderedContacts:
                self.orderedContacts[attendee_uid] = \
                    {'attendee': False,
                     'excused': False,
                     'absent': False,
                     'signer': False,
                     'signature_number': None,
                     'replacement': None,
                     'voter': False}
            self.orderedContacts[attendee_uid][attendee_type] = True

        for signatory_uid, signature_number in signatories.items():
            self.orderedContacts[signatory_uid]['signer'] = True
            self.orderedContacts[signatory_uid]['signature_number'] = signature_number

        for replaced_uid, replacer_uid in replacements.items():
            self.orderedContacts[replaced_uid]['replacement'] = replacer_uid

        for voter_uid in voters:
            self.orderedContacts[voter_uid]['voter'] = True

        self._p_changed = True

    security.declarePrivate('updateContacts')

    def updateContacts(self):
        '''After a meeting has been created or edited, we update here the info
           related to contacts implied in the meeting: attendees, excused,
           absents, signatories, replacements, ...'''
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self)

        if not cfg.isUsingContacts():
            return

        # attendees, excused, absents
        meeting_attendees = self.REQUEST.get('meeting_attendees', [])
        # remove leading muser_ and return a list of tuples, position_uid, attendee_type
        attendees = OrderedDict()
        for key in meeting_attendees:
            # remove leading muser_
            prefix, position_uid, attendee_type = key.split('_')
            attendees[position_uid] = attendee_type

        # signatories, remove ''
        meeting_signatories = [
            signatory for signatory in self.REQUEST.get('meeting_signatories', []) if signatory]
        signatories = {}
        for key in meeting_signatories:
            signatory, signature_number = key.split('__signaturenumber__')
            signatories[signatory] = signature_number

        # replacements, remove ''
        meeting_replacements = [
            replacer for replacer in self.REQUEST.get('meeting_replacements', []) if replacer]
        replacements = {}
        for key in meeting_replacements:
            replaced, replacer = key.split('__replacedby__')
            replacements[replaced] = replacer

        # voters
        meeting_voters = self.REQUEST.get('meeting_voters', [])
        # remove leading muser_ and return a list of tuples, position_uid, attendee_type
        voters = []
        for key in meeting_voters:
            # remove leading muser_ and ending _voter
            prefix, position_uid, suffix = key.split('_')
            voters.append(position_uid)

        self._doUpdateContacts(attendees, signatories, replacements, voters)

    security.declarePrivate('at_post_create_script')

    def at_post_create_script(self):
        ''' '''
        # place to store item absents
        self.itemAbsents = PersistentMapping()
        # place to store item excused
        self.itemExcused = PersistentMapping()
        # place to store item non attendees
        self.itemNonAttendees = PersistentMapping()
        # place to store item signatories
        self.itemSignatories = PersistentMapping()
        # place to store item votes
        self.itemVotes = PersistentMapping()
        # place to store attendees when using contacts
        self.orderedContacts = OrderedDict()
        self.updateTitle()
        self.updatePlace()
        self.computeDates()
        # Update contact-related info (attendees, signatories, replacements...)
        self.updateContacts()
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self)
        self.setMeetingConfigVersion(cfg.getConfigVersion())
        # addRecurringItemsIfRelevant(self, '_init_')
        # Apply potential transformations to richtext fields
        transformAllRichTextFields(self)
        # Make sure we have 'text/html' for every Rich fields
        forceHTMLContentTypeForEmptyRichFields(self)
        self.update_local_roles()
        # activate the faceted navigation
        enableFacetedDashboardFor(self,
                                  xmlpath=os.path.dirname(__file__) +
                                  '/faceted_conf/default_dashboard_widgets.xml')
        self.setLayout('meeting_view')
        # update every items itemReference if needed
        self.updateItemReferences(check_needed=True)
        # invalidate last meeting modified
        invalidate_cachekey_volatile_for('Products.PloneMeeting.Meeting.modified', get_again=True)
        # Call sub-product-specific behaviour
        self.adapted().onEdit(isCreated=True)
        self.reindexObject()

    def _update_after_edit(self):
        """Convenience method that make sure ObjectModifiedEvent and
           at_post_edit_script are called, like it is the case in
           Archetypes.BaseObject.processForm.
           We also call reindexObject here so we avoid multiple reindexation
           as it is already done in processForm.
           This is called when we change something on a meeting and we do not
           use processForm."""
        # WARNING, we do things the same order processForm do it
        # reindexObject is done in _processForm, then notify and
        # call to at_post_edit_script are done
        self.reindexObject()
        notify(ObjectEditedEvent(self))
        self.at_post_edit_script()

    security.declarePrivate('at_post_edit_script')

    def at_post_edit_script(self):
        '''Updates the meeting title.'''
        need_reindex = self.updateTitle()
        need_reindex = self.updatePlace() or need_reindex
        # Update contact-related info (attendees, signatories, replacements...)
        self.updateContacts()
        # Add a line in history if historized fields have changed
        addDataChange(self)
        # Apply potential transformations to richtext fields
        transformAllRichTextFields(self)
        # Make sure we have 'text/html' for every Rich fields
        forceHTMLContentTypeForEmptyRichFields(self)
        # update every items itemReference if needed
        self.updateItemReferences(check_needed=True)
        # update local roles as power observers local roles may vary depending on meeting_access_on
        self.update_local_roles()
        # Call sub-product-specific behaviour
        self.adapted().onEdit(isCreated=False)
        # invalidate last meeting modified
        invalidate_cachekey_volatile_for(
            'Products.PloneMeeting.Meeting.modified', get_again=True)
        # invalidate item voters vocabulary in case new voters (un)selected
        invalidate_cachekey_volatile_for(
            'Products.PloneMeeting.vocabularies.itemvotersvocabulary', get_again=True)
        # invalidate assembly async load on meeting
        invalidate_cachekey_volatile_for(
            'Products.PloneMeeting.browser.async.AsyncLoadMeetingAssemblyAndSignatures',
            get_again=True)
        if need_reindex:
            self.reindexObject()

    def update_local_roles(self, **kwargs):
        """Update various local roles."""
        # remove every localRoles then recompute
        old_local_roles = self.__ac_local_roles__.copy()
        self.__ac_local_roles__.clear()
        # add 'Owner' local role
        self.manage_addLocalRoles(self.owner_info()['id'], ('Owner',))
        # Update every 'power observers' local roles given to the
        # corresponding MeetingConfig.powerObsevers
        # it is done on every edit because of 'meeting_access_on' TAL expression
        self._updatePowerObserversLocalRoles()
        _addManagedPermissions(self)
        # notify that localRoles have been updated
        notify(MeetingLocalRolesUpdatedEvent(self, old_local_roles))
        # not really necessary here but easier
        # update annexes categorized_elements to store 'visible_for_groups'
        updateAnnexesAccess(self)
        # reindex object security except if avoid_reindex=True and localroles are the same
        avoid_reindex = kwargs.get('avoid_reindex', False)
        if not avoid_reindex or old_local_roles != self.__ac_local_roles__:
            self.reindexObjectSecurity()

    def _updatePowerObserversLocalRoles(self):
        '''Give local roles to the groups defined in MeetingConfig.powerObservers.'''
        extra_expr_ctx = _base_extra_expr_ctx(self)
        extra_expr_ctx.update({'meeting': self, })
        cfg = extra_expr_ctx['cfg']
        cfg_id = cfg.getId()
        meetingState = self.query_state()
        for po_infos in cfg.getPowerObservers():
            if meetingState in po_infos['meeting_states'] and \
               _evaluateExpression(self,
                                   expression=po_infos['meeting_access_on'],
                                   extra_expr_ctx=extra_expr_ctx):
                powerObserversGroupId = "%s_%s" % (cfg_id, po_infos['row_id'])
                self.manage_addLocalRoles(powerObserversGroupId, (READER_USECASES['powerobservers'],))

    security.declareProtected(ModifyPortalContent, 'onEdit')

    def onEdit(self, isCreated):
        '''See doc in interfaces.py.'''
        pass

    security.declarePublic('wfConditions')

    def wfConditions(self):
        '''Returns the adapter that implements the interface that proposes
           methods for use as conditions in the workflow associated with this
           meeting.'''
        return getWorkflowAdapter(self, conditions=True)

    security.declarePublic('wfActions')

    def wfActions(self):
        '''Returns the adapter that implements the interface that proposes
           methods for use as actions in the workflow associated with this
           meeting.'''
        return getWorkflowAdapter(self, conditions=False)

    security.declarePublic('adapted')

    def adapted(self):
        '''Gets the "adapted" version of myself. If no custom adapter is found,
           this method returns me.'''
        return getCustomAdapter(self)

    security.declarePublic('hasHistory')

    def hasHistory(self, fieldName=None):
        '''See doc in utils.py.'''
        return hasHistory(self, fieldName)

    security.declarePublic('attribute_is_used')

    def attribute_is_used(self, name):
        '''Is the attribute named p_name used in this meeting config ?'''
        tool = api.portal.get_tool('portal_plonemeeting')
        meetingConfig = tool.getMeetingConfig(self)
        return (name in meetingConfig.getUsedMeetingAttributes())

    def query_state_cachekey(method, self):
        '''cachekey method for self.query_state.'''
        return self.workflow_history

    security.declarePublic('query_state')

    @ram.cache(query_state_cachekey)
    def query_state(self):
        '''In what state am I ?'''
        wfTool = api.portal.get_tool('portal_workflow')
        return wfTool.getInfoFor(self, 'review_state')

    security.declarePublic('getSelf')

    def getSelf(self):
        '''Similar to MeetingItem.getSelf. Check MeetingItem.py for more
           info.'''
        res = self
        if self.getTagName() != 'Meeting':
            res = self.context
        return res

    security.declarePublic('setFieldFromAjax')

    def setFieldFromAjax(self, fieldName, fieldValue):
        '''See doc in utils.py.'''
        return set_field_from_ajax(self, fieldName, fieldValue)

    security.declarePublic('getFieldVersion')

    def getFieldVersion(self, fieldName, changes=False):
        '''See doc in utils.py.'''
        return getFieldVersion(self, fieldName, changes)

    security.declarePublic('isDecided')

    def isDecided(self):
        meeting = self.getSelf()
        return meeting.query_state() in ('decided', 'closed', 'decisions_published', )

    security.declarePublic('getSpecificMailContext')

    def getSpecificMailContext(self, event, translationMapping):
        '''See doc in interfaces.py.'''
        return None

    security.declarePublic('includeMailRecipient')

    def includeMailRecipient(self, event, userId):
        '''See doc in interfaces.py.'''
        return True

    security.declarePrivate('addRecurringItems')

    def addRecurringItems(self, recurringItems):
        '''Inserts into this meeting some p_recurringItems. The newly created
           items are copied from recurring items (contained in the meeting
           config) to the folder containing this meeting.'''
        destFolder = self.getParentNode()
        newItems = []
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self)
        for recurringItem in recurringItems:
            newItems.append(recurringItem.clone(cloneEventAction='Add recurring item',
                                                destFolder=destFolder,
                                                keepProposingGroup=True,
                                                newPortalType=cfg.getItemTypeName()))
        for newItem in newItems:
            # Put the new item in the correct state
            adap = newItem.adapted()
            error = adap.addRecurringItemToMeeting(self)
            if not error:
                notify(ItemDuplicatedFromConfigEvent(newItem, 'as_recurring_item'))
                newItem.reindexObject()

    security.declarePublic('fieldIsEmpty')

    def fieldIsEmpty(self, name):
        '''Is field named p_name empty ?'''
        return fieldIsEmpty(name, self)

    security.declarePublic('numberOfItems')

    def numberOfItems(self):
        '''How much items in this meeting ?'''
        return str(len(self.getRawItems()))

    security.declarePrivate('manage_beforeDelete')

    def manage_beforeDelete(self, item, container):
        '''This is a workaround to avoid a Plone design problem where it is
           possible to remove a folder containing objects you can not remove.'''
        # bypassed for migration to DX
        return
        # If we are here, everything has already been checked before.
        # Just check that the meeting is myself or a Plone Site.
        # We can remove an meeting directly but not "through" his container.
        if item.meta_type not in ('Plone Site', 'Meeting'):
            user = api.user.get_current()
            logger.warn(BEFOREDELETE_ERROR % (user.getId(), self.id))
            raise BeforeDeleteException("can_not_delete_meeting_container")
        # we are removing the meeting
        if item.meta_type == 'Meeting':
            member = api.user.get_current()
            if member.has_role('Manager'):
                item.REQUEST.set('items_to_remove', item.get_items())
        OrderedBaseFolder.manage_beforeDelete(self, item, container)

    security.declarePublic('showAttendeesFields')

    def showAttendeesFields(self):
        '''Display attendee related fields in view/edit?'''
        return (self.attribute_is_used('attendees') or self.get_attendees()) and not self.getAssembly()

    def shownAssemblyFields_cachekey(method, self):
        '''cachekey method for self.shownAssemblyFields.'''
        # do only re-compute if cfg changed or self.changed
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self)
        return (cfg.getId(), cfg._p_mtime, self.modified())

    security.declarePublic('shownAssemblyFields')

    @ram.cache(shownAssemblyFields_cachekey)
    def shownAssemblyFields(self):
        '''Return the list of shown assembly field :
           - used assembly fields;
           - not empty assembly fields.'''
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self)
        usedAttrs = cfg.getUsedMeetingAttributes()
        # get assembly fields
        fields = cfg._assembly_fields(field_name=False)
        return [field.getName() for field in fields
                if field.getName() in usedAttrs or field.get(self)]

    security.declarePublic('showSignatures')

    def showSignatures(self):
        '''Show the 'signatures' field?'''
        return self.attribute_is_used('signatures') or self.getSignatures()

    security.declarePublic('show_votesObservations')

    def show_votesObservations(self):
        '''See doc in interfaces.py.'''
        meeting = self.getSelf()
        tool = api.portal.get_tool('portal_plonemeeting')
        res = tool.isManager(meeting)
        if not res:
            cfg = tool.getMeetingConfig(meeting)
            res = isPowerObserverForCfg(cfg) or \
                meeting.adapted().isDecided()
        return res

    security.declarePublic('show_votes')

    def show_votes(self):
        '''See doc in interfaces.py.'''
        res = False
        meeting = self.getSelf()
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(meeting)
        if cfg.getUseVotes() or meeting.get_voters():
            res = True
        return res

    security.declarePublic('getPreviousMeeting')

    def getPreviousMeeting(self, searchMeetingsInterval=60):
        '''Gets the previous meeting based on meeting date. We only search among
           meetings in the previous p_searchMeetingsInterval, which is a number
           of days. If no meeting is found, the method returns None.'''
        meetingDate = self.getDate()
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self)
        meetingTypeName = cfg.getMeetingTypeName()
        catalog = api.portal.get_tool('portal_catalog')
        # find every meetings before searchMeetingsInterval days before self
        brains = catalog(portal_type=meetingTypeName,
                         getDate={'query': meetingDate - searchMeetingsInterval,
                                  'range': 'min'}, sort_on='getDate',
                         sort_order='reverse')
        res = None
        for brain in brains:
            if brain.getDate < meetingDate:
                res = brain
                break
        if res:
            res = res.getObject()
        return res

    security.declarePublic('getNextMeeting')

    def getNextMeeting(self, cfgId='', dateGap=0):
        '''Gets the next meeting based on meeting date.
           p_cfg can be used to compare meetings from another meetingconfig
           with meeting from the current config.
           p_dateGap is the number of 'dead days' following the date of
           the current meeting in which we do not look for next meeting'''
        tool = api.portal.get_tool('portal_plonemeeting')
        if not cfgId:
            cfg = tool.getMeetingConfig(self)
        else:
            cfg = getattr(tool, cfgId)
        return get_next_meeting(meetingDate=self.getDate(), cfg=cfg, dateGap=dateGap)

    security.declareProtected(ModifyPortalContent, 'processForm')

    def processForm(self, *args, **kwargs):
        '''We override this method because we may need to remember previous
           values of historized fields.'''
        if not self.isTemporary():
            self._v_previousData = rememberPreviousData(self)
        return OrderedBaseFolder.processForm(self, *args, **kwargs)

    security.declarePublic('showInsertOrRemoveSelectedItemsAction')

    def showInsertOrRemoveSelectedItemsAction(self):
        '''See doc in interfaces.py.'''
        meeting = self.getSelf()
        member = api.user.get_current()
        return bool(member.has_permission(ModifyPortalContent, meeting) and
                    not meeting.query_state() in meeting.meetingClosedStates)

    security.declarePublic('getLabelAssembly')

    def getLabelAssembly(self):
        '''
          Depending on the fact that we use 'assembly' alone or
          'assembly, excused, absents', we will translate the 'assembly' label
          a different way.
        '''
        tool = api.portal.get_tool('portal_plonemeeting')
        cfg = tool.getMeetingConfig(self)
        usedMeetingAttributes = cfg.getUsedMeetingAttributes()
        if 'assemblyExcused' in usedMeetingAttributes or \
           'assemblyAbsents' in usedMeetingAttributes:
            return _('PloneMeeting_label_attendees')
        else:
            return _('meeting_assembly')


registerType(Meeting, PROJECTNAME)
